from typing import TYPE_CHECKING, Iterator, List, Optional

from pyminion.core import Card
from pyminion.decider import Decider
from pyminion.player import Player

if TYPE_CHECKING:
    from pyminion.game import Game


class BotDecider:
    """
    Basic representation of Bot decision making.
    These methods can be implemented with specific game logic
    when creating new bots. In this class, these methods just return
    a valid response as to not crash the game.

    Does not implement any logic for playing or buying cards, as those functions
    (action_priority and buy_priority), are meant to be implemented when defining new bots.

    Does implement default responses to any decision the bot might have to make
    when playing a card or responding to an attack as to not crash the game.

    """

    def action_priority(self, player: "Player", game: "Game") -> Iterator[Card]:
        """
        Add logic for playing action cards through this method

        This function should be a generator where each call
        yields a desired card to play if conditions are met

        """
        raise NotImplementedError("action_priority is not implemented")

    def action_phase_decision(
        self,
        valid_actions: List["Card"],
        player: "Player",
        game: "Game",
    ) -> Optional["Card"]:
        for card in self.action_priority(player, game):
            if card in player.hand.cards:
                return card

        return None

    def treasure_phase_decision(
        self,
        valid_treasures: List["Card"],
        player: "Player",
        game: "Game",
    ) -> List["Card"]:
        return valid_treasures

    def buy_priority(self, player: "Player", game: "Game") -> Iterator[Card]:
        """
        Add logic for buy priority through this method

        This function should be a generator where each call
        yields a desired card to buy if conditions are met

        """
        raise NotImplementedError("buy_priority is not implemented")

    def buy_phase_decision(
        self,
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
    ) -> Optional["Card"]:
        for card in self.buy_priority(player, game):
            if game.supply.pile_length(card.name) > 0:
                return card

        return None

    def binary_decision(
        self,
        prompt: str,
        card: Card,
        player: "Player",
        game: "Game",
        relevant_cards: Optional[List[Card]] = None,
    ) -> bool:
        return True

    def multiple_option_decision(
        self,
        card: "Card",
        options: List[str],
        player: "Player",
        game: "Game",
        num_choices: int = 1,
        unique: bool = True,
    ) -> List[int]:
        return list(range(num_choices))

    def discard_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_discard: int = 0,
        max_num_discard: int = -1,
    ) -> List["Card"]:
        return valid_cards[:min_num_discard]

    def trash_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_trash: int = 0,
        max_num_trash: int = -1,
    ) -> List["Card"]:
        return valid_cards[:min_num_trash]

    def gain_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_gain: int = 0,
        max_num_gain: int = -1,
    ) -> List["Card"]:
        return valid_cards[:min_num_gain]

    def topdeck_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_topdeck: int = 0,
        max_num_topdeck: int = -1,
    ) -> List["Card"]:
        return valid_cards[:min_num_topdeck]

    def deck_position_decision(
        self,
        prompt: str,
        card: "Card",
        player: "Player",
        game: "Game",
        num_deck_cards: int,
    ) -> int:
        return num_deck_cards

    def reveal_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_reveal: int = 0,
        max_num_reveal: int = -1,
    ) -> List["Card"]:
        return valid_cards[:min_num_reveal]

    def pass_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_pass: int = 0,
        max_num_pass: int = -1,
    ) -> List["Card"]:
        return valid_cards[:min_num_pass]

    def name_card_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_name: int = 0,
        max_num_name: int = -1,
    ) -> List["Card"]:
        return valid_cards[:min_num_name]

    def multi_play_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        required: bool = True,
    ) -> Optional["Card"]:
        if required:
            return valid_cards[0]
        else:
            return None


class Bot(Player):
    """
    Abstract representation of Bot.

    """

    def __init__(
        self,
        decider: Optional[Decider] = None,
        player_id: str = "bot",
    ):
        decider = decider if decider else BotDecider()
        super().__init__(decider=decider, player_id=player_id)



import logging
import math
from typing import TYPE_CHECKING, List, Tuple

from pyminion.core import AbstractDeck, CardType, Action, Card, Treasure, Victory
from pyminion.exceptions import EmptyPile
from pyminion.player import Player

if TYPE_CHECKING:
    from pyminion.game import Game


logger = logging.getLogger()


class Copper(Treasure):
    def __init__(
        self,
        name: str = "Copper",
        cost: int = 0,
        type: Tuple[CardType] = (CardType.Treasure,),
        money: int = 1,
    ):
        super().__init__(name, cost, type, money)

    def play(self, player: Player, game: "Game"):
        player.playmat.add(self)
        player.hand.remove(self)
        player.state.money += self.money


class Silver(Treasure):
    def __init__(
        self,
        name: str = "Silver",
        cost: int = 3,
        type: Tuple[CardType] = (CardType.Treasure,),
        money: int = 2,
    ):
        super().__init__(name, cost, type, money)

    def play(self, player: Player, game: "Game"):

        # check if this is the first silver played and if there are any merchants in play
        if self not in player.playmat.cards:
            if num_merchants := len(
                [card for card in player.playmat.cards if card.name == "Merchant"]
            ):
                player.state.money += num_merchants

        player.playmat.add(self)
        player.hand.remove(self)
        player.state.money += self.money


class Gold(Treasure):
    def __init__(
        self,
        name: str = "Gold",
        cost: int = 6,
        type: Tuple[CardType] = (CardType.Treasure,),
        money: int = 3,
    ):
        super().__init__(name, cost, type, money)

    def play(self, player: Player, game: "Game"):
        player.playmat.add(self)
        player.hand.remove(self)
        player.state.money += self.money


class Estate(Victory):
    def __init__(
        self,
        name: str = "Estate",
        cost: int = 2,
        type: Tuple[CardType] = (CardType.Victory,),
    ):
        super().__init__(name, cost, type)

    def score(self, player: Player) -> int:
        vp = 1
        return vp


class Duchy(Victory):
    def __init__(
        self,
        name: str = "Duchy",
        cost: int = 5,
        type: Tuple[CardType] = (CardType.Victory,),
    ):
        super().__init__(name, cost, type)

    def score(self, player: Player) -> int:
        vp = 3
        return vp


class Province(Victory):
    def __init__(
        self,
        name: str = "Province",
        cost: int = 8,
        type: Tuple[CardType] = (CardType.Victory,),
    ):
        super().__init__(name, cost, type)

    def score(self, player: Player) -> int:
        vp = 6
        return vp


class Curse(Victory):
    def __init__(
        self,
        name: str = "Curse",
        cost: int = 0,
        type: Tuple[CardType] = (CardType.Curse,),
    ):
        super().__init__(name, cost, type)

    def score(self, player: Player) -> int:
        vp = -1
        return vp


class Gardens(Victory):
    """
    Worth 1VP for every 10 cards you have (round down)

    """

    def __init__(
        self,
        name: str = "Gardens",
        cost: int = 4,
        type: Tuple[CardType] = (CardType.Victory,),
    ):
        super().__init__(name, cost, type)

    def score(self, player: Player) -> int:
        total_count = len(player.get_all_cards())
        vp = math.floor(total_count / 10)
        return vp


class Smithy(Action):
    """
    + 3 cards

    """

    def __init__(
        self,
        name: str = "Smithy",
        cost: int = 4,
        type: Tuple[CardType] = (CardType.Action,),
        draw: int = 3,
    ):
        super().__init__(name, cost, type, draw=draw)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(3)


class Village(Action):
    """
    + 1 card, + 2 actions

    """

    def __init__(
        self,
        name: str = "Village",
        cost: int = 3,
        type: Tuple[CardType] = (CardType.Action,),
        actions: int = 2,
        draw: int = 1,
    ):
        super().__init__(name, cost, type, actions=actions, draw=draw)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 2
        player.draw()


class Laboratory(Action):
    """
    +2 cards, +1 action

    """

    def __init__(
        self,
        name: str = "Laboratory",
        cost: int = 5,
        type: Tuple[CardType] = (CardType.Action,),
        actions: int = 1,
        draw: int = 2,
    ):
        super().__init__(name, cost, type, actions=actions, draw=draw)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 1
        player.draw(2)


class Market(Action):
    """
    +1 card, +1 action, +1 money, +1 buy

    """

    def __init__(
        self,
        name: str = "Market",
        cost: int = 5,
        type: Tuple[CardType] = (CardType.Action,),
        actions: int = 1,
        draw: int = 1,
        money: int = 1,
    ):
        super().__init__(name, cost, type, actions, draw, money)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 1
        player.draw()
        player.state.money += 1
        player.state.buys += 1


class Moneylender(Action):
    """
    You may trash a copper from your hand for + 3 money

    """

    def __init__(
        self,
        name: str = "Moneylender",
        cost: int = 4,
        type: Tuple[CardType] = (CardType.Action,),
    ):
        super().__init__(name, cost, type)

    def play(
        self,
        player: Player,
        game: "Game",
        generic_play: bool = True,
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        if copper not in player.hand.cards:
            return

        response = player.decider.binary_decision(
            prompt="Do you want to trash a copper from your hand for +3 money? y/n: ",
            card=self,
            player=player,
            game=game,
        )

        if response:
            player.trash(target_card=copper, trash=game.trash)
            player.state.money += 3


class Cellar(Action):
    """
    +1 Action

    Discard any number of cards, then draw that many

    """

    def __init__(
        self,
        name: str = "Cellar",
        cost: int = 2,
        type: Tuple[CardType] = (CardType.Action,),
        actions: int = 1,
    ):
        super().__init__(name, cost, type, actions=actions)

    def play(
        self,
        player: Player,
        game: "Game",
        generic_play: bool = True,
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 1

        if not player.hand.cards:
            return

        discard_cards = player.decider.discard_decision(
            prompt="Enter the cards you would like to discard separated by commas: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
        )

        for card in discard_cards:
            player.discard(card)
        player.draw(len(discard_cards))


class Chapel(Action):
    """
    Trash up to 4 cards from your hand

    """

    def __init__(
        self,
        name: str = "Chapel",
        cost: int = 2,
        type: Tuple[CardType] = (CardType.Action,),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")
        if generic_play:
            super().generic_play(player)

        if not player.hand.cards:
            return

        trash_cards = player.decider.trash_decision(
            prompt="Enter up to 4 cards you would like to trash from your hand: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
            min_num_trash=0,
            max_num_trash=4,
        )
        assert len(trash_cards) <= 4

        for card in trash_cards:
            player.trash(card, game.trash)


class Workshop(Action):
    """
    Gain a card costing up to 4 money

    """

    def __init__(
        self,
        name: str = "Workshop",
        cost: int = 3,
        type: Tuple[CardType] = (CardType.Action,),
    ):
        super().__init__(name, cost, type)

    def play(self, player: Player, game: "Game", generic_play: bool = True) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        gain_cards = player.decider.gain_decision(
            prompt="Gain a card costing up to 4 money: ",
            card=self,
            valid_cards=[
                card for card in game.supply.avaliable_cards() if card.get_cost(player, game) <= 4
            ],
            player=player,
            game=game,
            min_num_gain=1,
            max_num_gain=1,
        )
        assert len(gain_cards) == 1
        gain_card = gain_cards[0]
        assert gain_card.get_cost(player, game) <= 4

        player.gain(gain_card, game.supply)


class Festival(Action):
    """
    + 2 actions, + 1 buy, + 2 money

    """

    def __init__(
        self,
        name: str = "Festival",
        cost: int = 5,
        type: Tuple[CardType] = (CardType.Action,),
        actions: int = 2,
        money: int = 2,
    ):
        super().__init__(name, cost, type, actions=actions, money=money)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 2
        player.state.money += 2
        player.state.buys += 1


class Harbinger(Action):
    """
    +1 card, +1 action

    Look through your discard pile. You may put a card from it onto your deck

    """

    def __init__(
        self,
        name: str = "Harbinger",
        cost: int = 3,
        type: Tuple[CardType] = (CardType.Action,),
        actions: int = 1,
        draw: int = 1,
    ):
        super().__init__(name, cost, type, actions=actions, draw=draw)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 1
        player.draw()

        if not player.discard_pile:
            return

        topdeck_cards = player.decider.topdeck_decision(
            prompt="You may select a card from your discard pile to put onto your deck: ",
            card=self,
            valid_cards=player.discard_pile.cards,
            player=player,
            game=game,
            min_num_topdeck=0,
            max_num_topdeck=1,
        )

        if len(topdeck_cards) == 0:
            return

        assert len(topdeck_cards) == 1
        topdeck_card = topdeck_cards[0]
        player.deck.add(player.discard_pile.remove(topdeck_card))


class Vassal(Action):
    """
    +2 money

    Discard the top card of your deck. If it's an action card you may play it.

    """

    def __init__(
        self,
        name: str = "Vassal",
        cost: int = 3,
        type: Tuple[CardType] = (CardType.Action,),
        money: int = 2,
    ):
        super().__init__(name, cost, type, money=money)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.money += 2
        player.draw(destination=player.discard_pile, silent=True)

        if not player.discard_pile:
            return

        discard_card = player.discard_pile.cards[-1]

        logger.info(f"{player} discards {discard_card}")

        if CardType.Action not in discard_card.type:
            return

        decision = player.decider.binary_decision(
            prompt=f"You discarded {discard_card.name}, would you like to play it? (y/n):  ",
            card=self,
            player=player,
            game=game,
        )

        if not decision:
            return

        played_card = player.discard_pile.cards.pop()
        player.playmat.add(played_card)
        player.exact_play(card=player.playmat.cards[-1], game=game, generic_play=False)
        return


class Artisan(Action):
    """
    Gain a card to your hand costing up to 5 money.

    Put a card from your hand onto your deck

    """

    def __init__(
        self,
        name: str = "Artisan",
        cost: int = 6,
        type: Tuple[CardType] = (CardType.Action,),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        gain_cards = player.decider.gain_decision(
            prompt="Gain a card costing up to 5 money: ",
            card=self,
            valid_cards=[
                card for card in game.supply.avaliable_cards() if card.get_cost(player, game) <= 5
            ],
            player=player,
            game=game,
            min_num_gain=1,
            max_num_gain=1,
        )
        assert len(gain_cards) == 1
        gain_card = gain_cards[0]
        assert gain_card.get_cost(player, game) <= 5

        player.gain(card=gain_card, supply=game.supply, destination=player.hand)

        topdeck_cards = player.decider.topdeck_decision(
            prompt="Put a card from your hand onto your deck: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
            min_num_topdeck=1,
            max_num_topdeck=1,
        )
        assert len(topdeck_cards) == 1
        topdeck_card = topdeck_cards[0]

        for card in player.hand.cards:
            if card == topdeck_card:
                player.deck.add(player.hand.remove(card))
                return


class Poacher(Action):
    """
    +1 card, +1 action, + 1 money

    Discard a card per empty Supply pile

    """

    def __init__(
        self,
        name: str = "Poacher",
        cost: int = 4,
        type: Tuple[CardType] = (CardType.Action,),
        actions: int = 1,
        draw: int = 1,
        money: int = 1,
    ):
        super().__init__(name, cost, type, actions, draw, money)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw()
        player.state.actions += 1
        player.state.money += 1

        empty_piles = game.supply.num_empty_piles()

        if empty_piles == 0:
            return

        discard_num = min(empty_piles, len(player.hand))

        discard_cards = player.decider.discard_decision(
            prompt=f"Discard {discard_num} card(s) from your hand: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
            min_num_discard=discard_num,
            max_num_discard=discard_num,
        )
        assert len(discard_cards) == discard_num

        for discard_card in discard_cards:
            player.discard(discard_card)


class CouncilRoom(Action):
    """
    +4 cards, +1 buy

    Each other player draws a card

    """

    def __init__(
        self,
        name: str = "Council Room",
        cost: int = 5,
        type: Tuple[CardType] = (CardType.Action,),
        draw: int = 4,
    ):
        super().__init__(name, cost, type, draw=draw)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(4)
        player.state.buys += 1

        for p in game.players:
            if p is not player:
                p.draw()


class Witch(Action):
    """
    +2 cards

    Each other player gains a curse

    """

    def __init__(
        self,
        name: str = "Witch",
        cost: int = 5,
        type: Tuple[CardType] = (CardType.Action, CardType.Attack),
        draw: int = 2,
    ):
        super().__init__(name, cost, type, draw=draw)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(2)

        for opponent in game.players:
            if opponent is not player:
                if opponent.is_attacked(player=player, attack_card=self, game=game):

                    # attempt to gain a curse. if curse pile is empty, proceed
                    try:
                        opponent.gain(
                            card=curse,
                            supply=game.supply,
                        )
                    except EmptyPile:
                        pass


class Moat(Action):
    """
    +2 cards

    When another player plays an attack card, you may first
    reveal this from your hand, to be unaffected by it

    """

    def __init__(
        self,
        name: str = "Moat",
        cost: int = 2,
        type: Tuple[CardType] = (CardType.Action, CardType.Reaction),
        draw: int = 2,
    ):
        super().__init__(name, cost, type, draw=draw)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(2)


class Merchant(Action):
    """
    +1 card, +1 action

    The first time you play a Silver this turn, +1 money

    """

    def __init__(
        self,
        name: str = "Merchant",
        cost: int = 3,
        type: Tuple[CardType] = (CardType.Action,),
        actions: int = 1,
        draw: int = 1,
    ):
        super().__init__(name, cost, type, actions=actions, draw=draw)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(1)
        player.state.actions += 1


class Bandit(Action):
    """
    Gain a Gold. Each other player reveals the top 2 cards of their deck,
    trashes a revealed treasure other than Copper, and discards the rest

    """

    def __init__(
        self,
        name: str = "Bandit",
        cost: int = 5,
        type: Tuple[CardType] = (CardType.Action, CardType.Attack),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        # attempt to gain a gold. if gold pile is empty, proceed
        try:
            player.gain(card=gold, supply=game.supply)
        except EmptyPile:
            pass

        for opponent in game.players:
            if opponent is not player:
                if opponent.is_attacked(player=player, attack_card=self, game=game):

                    revealed_cards = AbstractDeck()
                    opponent.draw(num_cards=2, destination=revealed_cards, silent=True)

                    logger.info(f"{opponent} reveals {revealed_cards}")

                    trash_card = None
                    for card in revealed_cards.cards:
                        if card.name == "Silver":
                            trash_card = card
                        elif card.name == "Gold" and not trash_card:
                            trash_card = card
                        elif (
                            CardType.Treasure in card.type
                            and card.name != "Copper"
                            and not trash_card
                        ):
                            trash_card = card

                    if trash_card:
                        game.trash.add(revealed_cards.remove(trash_card))

                    revealed_cards.move_to(opponent.discard_pile)
                    del revealed_cards


class Bureaucrat(Action):
    """
    Gain a Silver onto your deck. Each other player reveals a victory card from
    their hand and puts it onto their deck (or reveals a hand with no victory cards)

    """

    def __init__(
        self,
        name: str = "Bureaucrat",
        cost: int = 4,
        type: Tuple[CardType] = (CardType.Action, CardType.Attack),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        # attempt to gain a silver. if silver pile is empty, proceed
        try:
            player.gain(card=silver, supply=game.supply, destination=player.deck)
        except EmptyPile:
            pass

        for opponent in game.players:
            if opponent is not player and opponent.is_attacked(
                player=player, attack_card=self, game=game
            ):

                victory_cards = []
                for card in opponent.hand.cards:
                    if CardType.Victory in card.type:
                        victory_cards.append(card)

                if not victory_cards:
                    logger.info(f"{opponent} reveals hand: {opponent.hand}")
                    continue

                topdeck_cards = opponent.decider.topdeck_decision(
                    prompt="You must topdeck a Victory card from your hand: ",
                    card=self,
                    valid_cards=victory_cards,
                    player=opponent,
                    game=game,
                    min_num_topdeck=1,
                    max_num_topdeck=1,
                )
                assert len(topdeck_cards) == 1
                topdeck_card = topdeck_cards[0]

                opponent.deck.add(opponent.hand.remove(topdeck_card))
                logger.info(f"{opponent} topdecks {topdeck_card}")


class ThroneRoom(Action):
    """
    You may play an Action card from your hand twice

    """

    def __init__(
        self,
        name: str = "Throne Room",
        cost: int = 4,
        type: Tuple[CardType] = (CardType.Action,),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        action_cards = [card for card in player.hand.cards if CardType.Action in card.type]

        if not action_cards:
            return

        dp_card = player.decider.multi_play_decision(
            prompt="You may play an action card from your hand twice: ",
            card=self,
            valid_cards=action_cards,
            player=player,
            game=game,
            required=True,
        )
        assert dp_card is not None

        for card in player.hand.cards:
            if card.name == dp_card.name:
                player.playmat.add(player.hand.remove(card))
                state = None
                for i in range(2):
                    state = player.multi_play(card=card, game=game, state=state, generic_play=False)
                return


class Remodel(Action):
    """
    Trash a card from your hand. Gain a card costing up to $2 more than it

    """

    def __init__(
        self,
        name: str = "Remodel",
        cost: int = 4,
        type: Tuple[CardType] = (CardType.Action,),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        trash_cards = player.decider.trash_decision(
            prompt="Trash a card form your hand: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
            min_num_trash=1,
            max_num_trash=1,
        )
        assert len(trash_cards) == 1
        trash_card = trash_cards[0]

        max_cost = trash_card.get_cost(player, game) + 2
        gain_cards = player.decider.gain_decision(
            prompt=f"Gain a card costing up to {max_cost} money: ",
            card=self,
            valid_cards=[
                card
                for card in game.supply.avaliable_cards()
                if card.get_cost(player, game) <= max_cost
            ],
            player=player,
            game=game,
            min_num_gain=1,
            max_num_gain=1,
        )
        assert len(gain_cards) == 1
        gain_card = gain_cards[0]
        assert gain_card.get_cost(player, game) <= max_cost

        player.trash(trash_card, trash=game.trash)
        player.gain(gain_card, game.supply)


class Mine(Action):
    """
    You may trash a Treasure from your hand. Gain a Treasure to your hand costing up to $3 more than it

    """

    def __init__(
        self,
        name: str = "Mine",
        cost: int = 5,
        type: Tuple[CardType] = (CardType.Action,),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        treasures = [card for card in player.hand.cards if CardType.Treasure in card.type]

        if not treasures:
            return

        trash_cards = player.decider.trash_decision(
            prompt="You may trash a Treasure from your hand: ",
            card=self,
            valid_cards=treasures,
            player=player,
            game=game,
            min_num_trash=0,
            max_num_trash=1,
        )

        if len(trash_cards) == 0:
            return

        assert len(trash_cards) == 1
        trash_card = trash_cards[0]

        max_cost = trash_card.get_cost(player, game) + 3
        gain_cards = player.decider.gain_decision(
            prompt=f"Gain a Treasure card costing up to {max_cost} money to your hand: ",
            card=self,
            valid_cards=[
                card
                for card in game.supply.avaliable_cards()
                if CardType.Treasure in card.type and card.get_cost(player, game) <= trash_card.get_cost(player, game) + 3
            ],
            player=player,
            game=game,
            min_num_gain=1,
            max_num_gain=1,
        )
        assert len(gain_cards) == 1
        gain_card = gain_cards[0]
        assert CardType.Treasure in gain_card.type
        assert gain_card.get_cost(player, game) <= max_cost

        player.trash(trash_card, trash=game.trash)
        player.gain(gain_card, game.supply, destination=player.hand)


class Militia(Action):
    """
    +2 Money

    Each other player discards down to 3 cards in hand

    """

    def __init__(
        self,
        name: str = "Militia",
        cost: int = 4,
        type: Tuple[CardType] = (CardType.Action, CardType.Attack),
        money: int = 2,
    ):
        super().__init__(name, cost, type, money=money)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.money += 2

        for opponent in game.players:
            if opponent is not player and opponent.is_attacked(
                player=player, attack_card=self, game=game
            ):

                num_discard = len(opponent.hand) - 3
                if num_discard <= 0:
                    continue

                discard_cards = opponent.decider.discard_decision(
                    prompt=f"You must discard {num_discard} card(s) from your hand: ",
                    card=self,
                    valid_cards=opponent.hand.cards,
                    player=opponent,
                    game=game,
                    min_num_discard=num_discard,
                    max_num_discard=num_discard,
                )
                assert len(discard_cards) == num_discard

                for card in discard_cards:
                    opponent.discard(target_card=card)


class Sentry(Action):
    """
    +1 card, +1 action

    Look at the top 2 cards of your deck. Trash and/or discard any number of them. Put the rest back on top in any order

    """

    def __init__(
        self,
        name: str = "Sentry",
        cost: int = 5,
        type: Tuple[CardType] = (CardType.Action,),
        actions: int = 1,
        draw: int = 1,
    ):
        super().__init__(name, cost, type, actions=actions, draw=draw)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw()
        player.state.actions += 1

        revealed = AbstractDeck()
        player.draw(num_cards=2, destination=revealed, silent=True)
        logger.info(f"{player} looks at {revealed}")

        trash_cards = player.decider.trash_decision(
            prompt="Enter the cards you would like to trash: ",
            card=self,
            valid_cards=revealed.cards,
            player=player,
            game=game,
            min_num_trash=0,
            max_num_trash=2,
        )

        for card in trash_cards:
            revealed.remove(card)

        discard_cards: List[Card] = []
        if len(revealed.cards) > 0:
            discard_cards = player.decider.discard_decision(
                prompt="Enter the cards you would like to discard: ",
                card=self,
                valid_cards=revealed.cards,
                player=player,
                game=game,
            )
            for card in discard_cards:
                revealed.remove(card)

        reorder = False
        if len(revealed.cards) == 2:
            logger.info(
                f"Current order: {revealed.cards[0]} (Top), {revealed.cards[1]} (Bottom)"
            )
            reorder = player.decider.binary_decision(
                prompt="Would you like to switch the order of the cards?",
                card=self,
                player=player,
                game=game,
            )

        for card in trash_cards:
            game.trash.add(card)
            logger.info(f"{player} trashes {card}")
        for card in discard_cards:
            player.discard_pile.add(card)
            logger.info(f"{player} discards {card}")
        if revealed.cards:
            if reorder:
                for card in revealed.cards:
                    player.deck.add(card)
            else:
                for card in reversed(revealed.cards):
                    player.deck.add(card)
            logger.info(f"{player} topdecks {len(revealed.cards)} cards")


class Library(Action):
    """
    Draw until you have 7 cards in hand, skipping any Action cards you choose to; set those aside, discarding them afterwards

    """

    def __init__(
        self,
        name: str = "Library",
        cost: int = 5,
        type: Tuple[CardType] = (CardType.Action,),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        set_aside = AbstractDeck()
        while len(player.hand) < 7:

            if len(player.deck) == 0 and len(player.discard_pile) == 0:
                return

            player.draw(num_cards=1, destination=set_aside)
            drawn_card = set_aside.cards[-1]

            if CardType.Action in drawn_card.type:
                should_skip = player.decider.binary_decision(
                    prompt=f"You drew {drawn_card}, would you like to skip it? y/n: ",
                    card=self,
                    player=player,
                    game=game,
                    relevant_cards=[drawn_card],
                )
                if not should_skip:
                    player.hand.add(set_aside.remove(drawn_card))

            else:
                player.hand.add(set_aside.remove(drawn_card))

        if set_aside.cards:
            logger.info(f"{player} discards {set_aside}")
            set_aside.move_to(destination=player.discard_pile)


copper = Copper()
silver = Silver()
gold = Gold()
estate = Estate()
duchy = Duchy()
province = Province()
curse = Curse()

artisan = Artisan()
bandit = Bandit()
bureaucrat = Bureaucrat()
cellar = Cellar()
chapel = Chapel()
council_room = CouncilRoom()
festival = Festival()
gardens = Gardens()
harbinger = Harbinger()
laboratory = Laboratory()
library = Library()
market = Market()
merchant = Merchant()
militia = Militia()
mine = Mine()
moat = Moat()
moneylender = Moneylender()
poacher = Poacher()
remodel = Remodel()
sentry = Sentry()
smithy = Smithy()
throne_room = ThroneRoom()
vassal = Vassal()
village = Village()
witch = Witch()
workshop = Workshop()


base_set: List[Card] = [
    artisan,
    bandit,
    bureaucrat,
    cellar,
    chapel,
    council_room,
    festival,
    gardens,
    harbinger,
    laboratory,
    library,
    market,
    merchant,
    militia,
    mine,
    moat,
    moneylender,
    poacher,
    remodel,
    sentry,
    smithy,
    throne_room,
    vassal,
    village,
    witch,
    workshop,
]



from typing import TYPE_CHECKING, Iterable, List, Literal, Optional, Tuple, Union, overload

from pyminion.bots.bot import Bot, BotDecider
from pyminion.core import CardType, Card, DeckCounter, Treasure, Victory, get_action_cards, get_treasure_cards, get_victory_cards
from pyminion.decider import Decider
from pyminion.exceptions import InvalidBotImplementation
from pyminion.expansions.base import duchy, estate, curse, gold, silver, copper
from pyminion.expansions.intrigue import Baron, Courtier, Lurker, Minion, Nobles, Pawn, Steward, Torturer
from pyminion.player import Player

if TYPE_CHECKING:
    from pyminion.game import Game


class OptimizedBotDecider(BotDecider):
    """
    Implements opinionated logic for playing and reacting to all cards in the base set.

    The intention is to inherit from this class to make concrete bot implementations.
    If inheriting from this bot, it is possible to change the way that a single card is executed
    by overwriting the card specific method at the bottom of this file.

    """

    @staticmethod
    def get_best_victory_card(valid_cards: List[Card], player: Player) -> Union[Victory, None]:
        """
        Get the victory card that has the highest score.

        """

        best_victory = max(
            get_victory_cards(valid_cards),
            key=lambda card: card.score(player),
            default=None,
        )
        return best_victory

    @staticmethod
    def sort_for_discard(cards: List[Card], actions: int, player: Player, game: "Game") -> List[Card]:
        """
        Sort list of cards from best discard candidate to worst discard candidate.
        First sort cards from lowest cost to highest cost. Then rearrange depending on remaining actions.
        If player has no remaining actions, prioritize discarding victory then action, then treasures.
        If player has remaining actions, prioritize discarding victory then treasure and action equally.

        """

        sorted_cards = sorted(cards, key=lambda card: card.get_cost(player, game))
        victory_cards = [
            card
            for card in sorted_cards
            if CardType.Victory in card.type or CardType.Curse in card.type
        ]
        non_victory_cards = [
            card
            for card in sorted_cards
            if CardType.Victory not in card.type and CardType.Curse not in card.type
        ]
        treasure_cards = [card for card in non_victory_cards if CardType.Treasure in card.type]
        action_cards = [
            card for card in non_victory_cards if CardType.Treasure not in card.type
        ]
        if actions == 0:
            return victory_cards + action_cards + treasure_cards
        else:
            return victory_cards + non_victory_cards

    @staticmethod
    def sort_for_trash(cards: Iterable[Card], player: Player, game: "Game") -> List[Card]:
        num_provinces = game.supply.pile_length(pile_name="Province")
        deck_money = player.get_deck_money()

        prioritized_cards: List[Tuple[int, Card]] = []
        for card in cards:
            if CardType.Curse in card.type:
                priority = 1
            elif card.name == "Estate" and num_provinces >= 5:
                priority = 2
            elif card.name == "Copper" and deck_money > 3:
                priority = 3
                deck_money -= 1
            else:
                priority = 100 + card.get_cost(player, game)

            prioritized_cards.append((priority, card))

        prioritized_cards.sort(key=lambda x: x[0])
        trash_cards = [x[1] for x in prioritized_cards]
        return trash_cards

    @staticmethod
    def get_optional_discard(cards: List[Card], player: Player) -> List[Card]:
        discard_cards: List[Card] = []
        actions = player.state.actions
        for card in cards:
            if CardType.Treasure in card.type:
                continue
            elif CardType.Curse in card.type:
                discard_cards.append(card)
            elif CardType.Victory in card.type and CardType.Action not in card.type:
                discard_cards.append(card)
            elif actions == 0 and CardType.Action in card.type:
                discard_cards.append(card)

        return discard_cards

    def determine_trash_cards(
        self, valid_cards: List[Card], player: Player, game: "Game"
    ) -> List[Card]:
        """
        Determine which cards should be trashed:

        Always trash Curse
        Trash Estate if number of provinces in supply >= 5
        Trash Copper if money in deck > 3 (keep enough to buy silver)
        Finally, sort the cards as to prioritize trashing estate over copper

        """

        deck_money = player.get_deck_money()
        trash_cards = []
        for card in valid_cards:
            if CardType.Curse in card.type:
                trash_cards.append(card)
            elif (
                card.name == "Estate"
                and game.supply.pile_length(pile_name="Province") >= 5
            ):
                trash_cards.append(card)
            elif card.name == "Copper" and deck_money > 3:
                trash_cards.append(card)
                deck_money -= 1

        sorted_trash_cards = self.sort_for_discard(
            cards=trash_cards,
            actions=1,
            player=player,
            game=game,
        )

        return sorted_trash_cards

    def binary_decision(
        self,
        prompt: str,
        card: Card,
        player: "Player",
        game: "Game",
        relevant_cards: Optional[List[Card]] = None,
    ) -> bool:
        if card.name == "Moneylender":
            return self.moneylender(player=player, game=game)
        elif card.name == "Vassal":
            return self.vassal(player=player, game=game, relevant_cards=relevant_cards)
        elif card.name == "Sentry":
            return self.sentry(player=player, game=game, relevant_cards=relevant_cards, binary=True)
        elif card.name == "Library":
            return self.library(player=player, game=game, relevant_cards=relevant_cards)
        elif card.name == "Moat":
            return self.moat(player=player, game=game, relevant_cards=relevant_cards)
        elif card.name == "Diplomat":
            return self.diplomat(player, game, binary=True)
        elif card.name == "Masquerade":
            return self.masquerade(player, game, binary=True)
        elif card.name == "Mill":
            return self.mill(player, game, binary=True)
        elif card.name == "Mining Village":
            return self.mining_village(player, game)
        else:
            return super().binary_decision(prompt, card, player, game, relevant_cards)

    def multiple_option_decision(
        self,
        card: "Card",
        options: List[str],
        player: "Player",
        game: "Game",
        num_choices: int = 1,
        unique: bool = True,
    ) -> List[int]:
        if card.name == "Baron":
            ret = self.baron(player, game)
            return [ret]
        elif card.name == "Courtier":
            return self.courtier(player, game, num_choices=num_choices, options=True)
        elif card.name == "Lurker":
            ret = self.lurker(player, game, options=True)
            return [ret]
        elif card.name == "Minion":
            ret = self.minion(player, game)
            return [ret]
        elif card.name == "Nobles":
            ret = self.nobles(player, game)
            return [ret]
        elif card.name == "Pawn":
            return self.pawn(player, game)
        elif card.name == "Steward":
            ret = self.steward(player, game, options=True)
            return [ret]
        elif card.name == "Torturer":
            ret = self.torturer(player, game, options=True)
            return [ret]
        else:
            return super().multiple_option_decision(card, options, player, game, num_choices, unique)

    def discard_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_discard: int = 0,
        max_num_discard: int = -1,
    ) -> List["Card"]:
        if card.name == "Cellar":
            return self.cellar(player=player, game=game, valid_cards=valid_cards)
        elif card.name == "Poacher":
            return self.poacher(player=player, game=game, valid_cards=valid_cards, num_discard=min_num_discard)
        elif card.name == "Militia":
            return self.militia(player=player, game=game, valid_cards=valid_cards, num_discard=min_num_discard)
        elif card.name == "Sentry":
            return self.sentry(player=player, game=game, valid_cards=valid_cards, discard=True)
        elif card.name == "Diplomat":
            return self.diplomat(player, game, valid_cards, min_num_discard, discard=True)
        elif card.name == "Mill":
            return self.mill(player, game, discard=True)
        elif card.name == "Torturer":
            return self.torturer(player, game, valid_cards=valid_cards, num_discard=min_num_discard, discard=True)
        else:
            return super().discard_decision(prompt, card, valid_cards, player, game, min_num_discard, max_num_discard)

    def trash_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_trash: int = 0,
        max_num_trash: int = -1,
    ) -> List["Card"]:
        if card.name == "Remodel":
            ret = self.remodel(player=player, game=game, valid_cards=valid_cards, trash=True)
            return [ret]
        elif card.name == "Mine":
            ret = self.mine(player=player, game=game, valid_cards=valid_cards, trash=True)
            return [ret]
        elif card.name == "Chapel":
            return self.chapel(player=player, game=game, valid_cards=valid_cards)
        elif card.name == "Sentry":
            return self.sentry(player=player, game=game, valid_cards=valid_cards, trash=True)
        elif card.name == "Lurker":
            ret = self.lurker(player, game, valid_cards, trash=True)
            return [ret]
        elif card.name == "Masquerade":
            ret = self.masquerade(player, game, valid_cards, trash=True)
            return [ret]
        elif card.name == "Replace":
            ret = self.replace(player, game, valid_cards, trash=True)
            return [ret]
        elif card.name == "Steward":
            return self.steward(player, game, valid_cards=valid_cards, trash=True)
        elif card.name == "Trading Post":
            return self.trading_post(player, game, valid_cards)
        elif card.name == "Upgrade":
            ret = self.upgrade(player, game, valid_cards, trash=True)
            return [ret]
        else:
            return super().trash_decision(prompt, card, valid_cards, player, game, min_num_trash, max_num_trash)

    def gain_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_gain: int = 0,
        max_num_gain: int = -1,
    ) -> List["Card"]:
        if card.name == "Artisan":
            ret = self.artisan(player=player, game=game, gain=True)
            return [ret]
        elif card.name == "Workshop":
            ret = self.workshop(player=player, game=game)
            return [ret]
        elif card.name == "Remodel":
            ret = self.remodel(player=player, game=game, valid_cards=valid_cards, gain=True)
            return [ret]
        elif card.name == "Mine":
            ret = self.mine(player=player, game=game, valid_cards=valid_cards, gain=True)
            return [ret]
        elif card.name == "Ironworks":
            ret = self.ironworks(player, game, valid_cards)
            return [ret]
        elif card.name == "Lurker":
            ret = self.lurker(player, game, valid_cards, gain=True)
            return [ret]
        elif card.name == "Replace":
            ret = self.replace(player, game, valid_cards, gain=True)
            return [ret]
        elif card.name == "Swindler":
            ret = self.swindler(player, game, valid_cards)
            return [ret]
        elif card.name == "Upgrade":
            ret = self.upgrade(player, game, valid_cards, gain=True)
            return [ret]
        else:
            return super().gain_decision(prompt, card, valid_cards, player, game, min_num_gain, max_num_gain)

    def topdeck_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_topdeck: int = 0,
        max_num_topdeck: int = -1,
    ) -> List["Card"]:
        if card.name == "Artisan":
            ret = self.artisan(player=player, game=game, valid_cards=valid_cards, topdeck=True)
            return [ret]
        elif card.name == "Harbinger":
            ret = self.harbinger(player=player, game=game, valid_cards=valid_cards)
            return [] if ret is None else [ret]
        elif card.name == "Bureaucrat":
            ret = self.bureaucrat(player=player, game=game, valid_cards=valid_cards)
            return [ret]
        elif card.name == "Courtyard":
            ret = self.courtyard(player, game, valid_cards)
            return [ret]
        elif card.name == "Patrol":
            return self.patrol(player, game, valid_cards)
        elif card.name == "Secret Passage":
            ret = self.secret_passage(player, game, valid_cards=valid_cards, topdeck=True)
            return [ret]
        else:
            return super().topdeck_decision(prompt, card, valid_cards, player, game, min_num_topdeck, max_num_topdeck)

    def deck_position_decision(
        self,
        prompt: str,
        card: "Card",
        player: "Player",
        game: "Game",
        num_deck_cards: int,
    ) -> int:
        if card.name == "Secret Passage":
            return self.secret_passage(player, game, num_deck_cards=num_deck_cards, pos=True)
        else:
            return super().deck_position_decision(prompt, card, player, game, num_deck_cards)

    def reveal_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_reveal: int = 0,
        max_num_reveal: int = -1,
    ) -> List["Card"]:
        if card.name == "Courtier":
            ret = self.courtier(player, game, valid_cards=valid_cards, reveal=True)
            return [ret]
        else:
            return super().reveal_decision(prompt, card, valid_cards, player, game, min_num_reveal, max_num_reveal)

    def pass_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_pass: int = 0,
        max_num_pass: int = -1,
    ) -> List["Card"]:
        if card.name == "Masquerade":
            ret = self.masquerade(player, game, valid_cards, pass_=True)
            return [ret]
        else:
            return super().pass_decision(prompt, card, valid_cards, player, game, min_num_pass, max_num_pass)

    def name_card_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_name: int = 0,
        max_num_name: int = -1,
    ) -> List["Card"]:
        if card.name == "Wishing Well":
            ret = self.wishing_well(player, game)
            return [ret]
        else:
            return super().name_card_decision(prompt, card, valid_cards, player, game, min_num_name, max_num_name)

    def multi_play_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        required: bool = True,
    ) -> Optional["Card"]:
        if card.name == "Throne Room":
            return self.throne_room(player=player, game=game, valid_cards=valid_cards)
        else:
            return super().multi_play_decision(prompt, card, valid_cards, player, game, required)

    # CARD SPECIFIC IMPLEMENTATIONS

    def moneylender(self, player: "Player", game: "Game") -> bool:
        return True

    def vassal(self, player: "Player", game: "Game", relevant_cards: Optional[List[Card]]) -> bool:
        return True

    @overload
    def sentry(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        relevant_cards: Optional[List[Card]] = None,
        trash: Literal[True] = True,
        discard: Literal[False] = False,
        binary: Literal[False] = False,
    ) -> List[Card]:
        ...

    @overload
    def sentry(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        relevant_cards: Optional[List[Card]] = None,
        trash: Literal[False] = False,
        discard: Literal[True] = True,
        binary: Literal[False] = False,
    ) -> List[Card]:
        ...

    @overload
    def sentry(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        relevant_cards: Optional[List[Card]] = None,
        trash: Literal[False] = False,
        discard: Literal[False] = False,
        binary: Literal[True] = True,
    ) -> bool:
        ...

    def sentry(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        relevant_cards: Optional[List[Card]] = None,
        trash: bool = False,
        discard: bool = False,
        binary: bool = False,
    ) -> Union[List[Card], bool]:
        if trash:
            if not valid_cards:
                return []
            return self.determine_trash_cards(
                valid_cards=valid_cards, player=player, game=game
            )
        if discard:
            if not valid_cards:
                return []
            return [
                card
                for card in valid_cards
                if card.name == "Copper"
                or CardType.Victory in card.type
                or CardType.Curse in card.type
            ]
        if binary:
            return False

        raise InvalidBotImplementation(
            "Either gain, topdeck or binary must be true when playing sentry"
        )

    def library(self, player: "Player", game: "Game", relevant_cards: Optional[List[Card]]) -> bool:
        if player.state.actions == 0:
            return True
        else:
            return False

    def cellar(self, player: "Player", game: "Game", valid_cards: List[Card]) -> List[Card]:
        return [
            card
            for card in valid_cards
            if card.name == "Copper" or CardType.Victory in card.type or CardType.Curse in card.type
        ]

    def moat(self, player: "Player", game: "Game", relevant_cards: Optional[List[Card]]) -> bool:
        return True

    def poacher(
        self, player: "Player", game: "Game", valid_cards: List[Card], num_discard: Optional[int]
    ) -> List[Card]:
        if not num_discard:
            return []
        discard_order = self.sort_for_discard(
            cards=valid_cards, actions=player.state.actions, player=player, game=game
        )
        return discard_order[:num_discard]

    def militia(
        self, player: "Player", game: "Game", valid_cards: List[Card], num_discard: Optional[int]
    ) -> List[Card]:
        if not num_discard:
            return []
        discard_order = self.sort_for_discard(
            cards=valid_cards, actions=player.state.actions, player=player, game=game
        )
        return discard_order[:num_discard]

    def remodel(
        self, player: "Player", game: "Game", valid_cards: List[Card], trash: bool = False, gain: bool = False
    ) -> Card:
        if trash:
            min_price_card = min(valid_cards, key=lambda card: card.get_cost(player, game))
            return min_price_card

        if gain:
            max_price_card = max(valid_cards, key=lambda card: card.get_cost(player, game))
            return max_price_card

        raise InvalidBotImplementation(
            "Either gain or trash must be true when playing remodel"
        )

    def mine(
        self, player: "Player", game: "Game", valid_cards: List[Card], trash: bool = False, gain: bool = False
    ) -> Card:
        if trash:
            min_price_card = min(valid_cards, key=lambda card: card.get_cost(player, game))
            return min_price_card

        if gain:
            max_price_card = max(valid_cards, key=lambda card: card.get_cost(player, game))
            return max_price_card

        raise InvalidBotImplementation(
            "Either gain or trash must be true when playing mine"
        )

    def chapel(self, player: "Player", game: "Game", valid_cards: List[Card]) -> List[Card]:
        trash_cards = self.determine_trash_cards(
            valid_cards=valid_cards, player=player, game=game
        )
        while len(trash_cards) > 4:
            trash_cards.pop()
        return trash_cards

    def artisan(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        gain: bool = False,
        topdeck: bool = False,
    ) -> Card:
        if topdeck:
            for card in player.hand.cards:
                if CardType.Action in card.type and player.state.actions == 0:
                    return card
            else:
                return player.hand.cards[-1]
        if gain:
            if game.supply.pile_length(pile_name="Province") < 5:
                return duchy
            else:
                return silver

        raise InvalidBotImplementation(
            "Either gain or topdeck must be true when playing artisan"
        )

    def workshop(
        self,
        player: "Player",
        game: "Game",
    ) -> Card:
        if game.supply.pile_length(pile_name="Province") < 3:
            return estate
        else:
            return silver

    def harbinger(
        self,
        player: "Player",
        game: "Game",
        valid_cards: List[Card],
    ) -> Optional[Card]:
        # Do not topdeck victory cards
        best_topdeck = [card for card in valid_cards if CardType.Victory not in card.type]
        if not best_topdeck:
            return None
        # Topdeck highest price card if price > 2
        max_price_card = max(best_topdeck, key=lambda card: card.get_cost(player, game))
        if max_price_card.get_cost(player, game) > 2:
            return max_price_card
        else:
            return None

    def bureaucrat(
        self,
        player: "Player",
        game: "Game",
        valid_cards: List[Card],
    ) -> Card:
        sorted_cards = sorted(valid_cards, key=lambda card: card.get_cost(player, game))
        return sorted_cards[0]

    def throne_room(self, player: "Player", game: "Game", valid_cards: List[Card]) -> Card:
        # Double play most expensive card
        max_price_card = max(valid_cards, key=lambda card: card.get_cost(player, game))
        return max_price_card

    def baron(
        self,
        player: "Player",
        game: "Game",
    ) -> int:
        return Baron.Choice.DiscardEstate

    @overload
    def courtier(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        num_choices: int = 0,
        reveal: Literal[True] = True,
        options: Literal[False] = False,
    ) -> Card:
        ...

    @overload
    def courtier(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        num_choices: int = 0,
        reveal: Literal[False] = False,
        options: Literal[True] = True,
    ) -> List[int]:
        ...

    def courtier(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        num_choices: int = 0,
        reveal: bool = False,
        options: bool = False,
    ) -> Union[Card, List[int]]:
        if reveal:
            assert valid_cards is not None
            # find the card with the most types
            num_types = 0
            reveal_card = valid_cards[0]
            for c in valid_cards:
                if len(c.type) > num_types:
                    num_types = len(c.type)
                    reveal_card = c
            return reveal_card
        elif options:
            assert num_choices > 0
            counter = DeckCounter(player.get_all_cards())
            gold_count = counter[gold]
            has_actions = any(CardType.Action in c.type for c in player.hand.cards)

            # prioritize choices
            choices: List[int] = []
            if has_actions:
                choices.append(Courtier.Choice.Action)
            if gold_count < 4:
                choices.append(Courtier.Choice.GainGold)
            choices.append(Courtier.Choice.Money)

            # add remaining choices
            for c in Courtier.Choice:
                if c.value not in choices:
                    choices.append(c.value)

            choices = choices[:num_choices]
            return choices
        else:
            raise InvalidBotImplementation(
                "Either reveal or options must be true when playing courier"
            )

    def courtyard(
        self,
        player: "Player",
        game: "Game",
        valid_cards: List[Card],
    ) -> Card:
        if player.state.actions == 0:
            for card in valid_cards:
                if CardType.Action in card.type:
                    return card

        return valid_cards[-1]

    @overload
    def diplomat(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Union[List[Card], None] = None,
        num_discard: int = 0,
        binary: Literal[True] = True,
        discard: Literal[False] = False,
    ) -> bool:
        ...

    @overload
    def diplomat(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Union[List[Card], None] = None,
        num_discard: int = 0,
        binary: Literal[False] = False,
        discard: Literal[True] = True,
    ) -> List[Card]:
        ...

    def diplomat(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Union[List[Card], None] = None,
        num_discard: int = 0,
        binary: bool = False,
        discard: bool = False,
    ) -> Union[bool, List[Card]]:
        if binary:
            return True
        elif discard:
            assert valid_cards is not None
            discard_order = self.sort_for_discard(
                cards=valid_cards, actions=player.state.actions, player=player, game=game
            )
            return discard_order[:num_discard]
        else:
            raise InvalidBotImplementation(
                "Either binary or discard must be true when playing diplomat"
            )

    def ironworks(
        self,
        player: "Player",
        game: "Game",
        valid_cards: List[Card],
    ) -> Card:
        best_victory = self.get_best_victory_card(valid_cards, player)
        if game.supply.pile_length(pile_name="Province") < 3 and best_victory is not None:
            return best_victory
        else:
            return silver

    @overload
    def lurker(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Union[List[Card], None] = None,
        options: Literal[True] = True,
        trash: Literal[False] = False,
        gain: Literal[False] = False,
    ) -> int:
        ...

    @overload
    def lurker(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Union[List[Card], None] = None,
        options: Literal[False] = False,
        trash: Literal[True] = True,
        gain: Literal[False] = False,
    ) -> Card:
        ...

    @overload
    def lurker(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Union[List[Card], None] = None,
        options: Literal[False] = False,
        trash: Literal[False] = False,
        gain: Literal[True] = True,
    ) -> Card:
        ...

    def lurker(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Union[List[Card], None] = None,
        options: bool = False,
        trash: bool = False,
        gain: bool = False,
    ) -> Union[int, Card]:
        if options:
            if any(CardType.Action in c.type for c in game.trash.cards):
                return Lurker.Choice.GainAction
            else:
                return Lurker.Choice.TrashAction
        elif trash or gain:
            assert valid_cards is not None
            max_price_card = max(valid_cards, key=lambda card: card.get_cost(player, game))
            return max_price_card
        else:
            raise InvalidBotImplementation(
                "Either options, trash, or gain must be true when playing lurker"
            )

    @overload
    def masquerade(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        pass_: Literal[True] = True,
        binary: Literal[False] = False,
        trash: Literal[False] = False,
    ) -> Card:
        ...

    @overload
    def masquerade(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        pass_: Literal[False] = False,
        binary: Literal[True] = True,
        trash: Literal[False] = False,
    ) -> bool:
        ...

    @overload
    def masquerade(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        pass_: Literal[False] = False,
        binary: Literal[False] = False,
        trash: Literal[True] = True,
    ) -> Card:
        ...

    def masquerade(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        pass_: bool = False,
        binary: bool = False,
        trash: bool = False,
    ) -> Union[bool, Card]:
        if pass_:
            assert valid_cards is not None
            cards = self.determine_trash_cards(valid_cards, player, game)
            return cards[0]
        elif binary:
            cards = self.determine_trash_cards(player.hand.cards, player, game)
            return len(cards) > 0
        elif trash:
            assert valid_cards is not None
            cards = self.determine_trash_cards(valid_cards, player, game)
            return cards[0]
        else:
            raise InvalidBotImplementation(
                "Either pass_, binary, or trash must be true when playing masquerade"
            )

    @overload
    def mill(
        self,
        player: "Player",
        game: "Game",
        binary: Literal[True] = True,
        discard: Literal[False] = False,
    ) -> bool:
        ...

    @overload
    def mill(
        self,
        player: "Player",
        game: "Game",
        binary: Literal[False] = False,
        discard: Literal[True] = True,
    ) -> List[Card]:
        ...

    def mill(
        self,
        player: "Player",
        game: "Game",
        binary: bool = False,
        discard: bool = False,
    ) -> Union[bool, List[Card]]:
        actions = player.state.actions
        cards = self.sort_for_discard(player.hand.cards, actions, player, game)
        cards = cards[:2]
        if binary:
            if len(cards) < 2:
                return False
            money = 0
            for card in get_treasure_cards(cards):
                money += card.money
            # discard cards if they provide less money than the
            # +$2 mill will give for discarding
            return money < 2
        elif discard:
            return cards
        else:
            raise InvalidBotImplementation(
                "Either binary or discard must be true when playing mill"
            )

    def mining_village(
        self,
        player: "Player",
        game: "Game",
    ) -> bool:
        return game.supply.pile_length("Province") < 3

    def minion(
        self,
        player: "Player",
        game: "Game",
    ) -> int:
        has_action_cards = False
        hand_money = 0
        for card in player.hand.cards:
            if CardType.Action in card.type:
                has_action_cards = True
            if CardType.Treasure in card.type:
                assert isinstance(card, Treasure)
                hand_money += card.money

        # don't discard hand if we can play more actions or if we have money in hand
        if (has_action_cards and player.state.actions > 0) or hand_money >= 3:
            return Minion.Choice.Money
        return Minion.Choice.DiscardDrawAttack

    def nobles(
        self,
        player: "Player",
        game: "Game",
    ) -> int:
        action_card_count = sum(1 for _ in get_action_cards(player.hand.cards))
        if action_card_count > player.state.actions:
            return Nobles.Choice.Actions
        return Nobles.Choice.Cards

    def patrol(
        self,
        player: "Player",
        game: "Game",
        valid_cards: List[Card],
    ) -> List[Card]:
        cards = sorted(
            valid_cards,
            key=lambda card: card.get_cost(player, game),
        )
        return cards

    def pawn(
        self,
        player: "Player",
        game: "Game",
    ) -> List[int]:
        action_card_count = 0
        total_money = player.state.money
        for card in player.hand.cards:
            if CardType.Action in card.type:
                action_card_count += 1
            if CardType.Treasure in card.type:
                assert isinstance(card, Treasure)
                total_money += card.money

        choices: List[int] = []
        if action_card_count > player.state.actions:
            choices.append(Pawn.Choice.Action)

        if total_money < 2:
            choices.append(Pawn.Choice.Money)
        elif total_money > 8:
            choices.append(Pawn.Choice.Buy)

        if len(choices) < 2:
            choices.append(Pawn.Choice.Card)
        if len(choices) < 2:
            choices.append(Pawn.Choice.Money)

        return choices

    def replace(
        self,
        player: "Player",
        game: "Game",
        valid_cards: List[Card],
        trash: bool = False,
        gain: bool = False,
    ) -> Card:
        num_provinces = game.supply.pile_length("Province")
        if trash:
            has_gold = False
            has_curse = False
            for card in valid_cards:
                if CardType.Curse in card.type:
                    has_curse = True
                elif card.name == "Gold":
                    has_gold = True

            if has_gold and num_provinces < 5:
                return gold

            if has_curse:
                return curse

            min_price_card = min(valid_cards, key=lambda card: card.get_cost(player, game))
            return min_price_card

        if gain:
            victory_cards = list(get_victory_cards(valid_cards))
            victory_cards.sort(key=lambda c: c.score(player), reverse=True)
            if len(victory_cards) > 0:
                victory_card = victory_cards[0]
                score = victory_card.score(player)
                if score >= 6:
                    return victory_card
                elif score >= 3 and num_provinces < 5:
                    return victory_card
                elif num_provinces < 3:
                    return victory_card

            max_price_card = max(valid_cards, key=lambda card: card.get_cost(player, game))
            return max_price_card

        raise InvalidBotImplementation(
            "Either trash or gain must be true when playing replace"
        )

    @overload
    def secret_passage(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        num_deck_cards: int = -1,
        topdeck: Literal[True] = True,
        pos: Literal[False] = False,
    ) -> Card:
        ...

    @overload
    def secret_passage(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        num_deck_cards: int = -1,
        topdeck: Literal[False] = False,
        pos: Literal[True] = True,
    ) -> int:
        ...

    def secret_passage(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        num_deck_cards: int = -1,
        topdeck: bool = False,
        pos: bool = False,
    ) -> Union[Card, int]:
        if topdeck:
            assert valid_cards is not None
            return valid_cards[-1]

        if pos:
            assert num_deck_cards >= 0
            return num_deck_cards # put the card on top of the deck

        raise InvalidBotImplementation(
            "Either topdeck or pos must be true when playing secret passage"
        )

    @overload
    def steward(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        options: Literal[True] = True,
        trash: Literal[False] = False,
    ) -> int:
        ...

    @overload
    def steward(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        options: Literal[False] = False,
        trash: Literal[True] = True,
    ) -> List[Card]:
        ...

    def steward(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        options: bool = False,
        trash: bool = False,
    ) -> Union[int, List[Card]]:
        if options:
            trash_cards = self.determine_trash_cards(player.hand.cards, player, game)
            if player.state.actions > 0:
                return Steward.Choice.Cards
            elif len(trash_cards) >= 2:
                return Steward.Choice.Trash
            else:
                return Steward.Choice.Money

        if trash:
            assert valid_cards is not None
            trash_cards = self.determine_trash_cards(valid_cards, player, game)
            return trash_cards[:2]

        raise InvalidBotImplementation(
            "Either options or trash must be true when playing steward"
        )

    def swindler(
        self,
        player: "Player",
        game: "Game",
        valid_cards: List[Card],
    ) -> Card:
        # prioritize cards
        prioritized_cards: List[Tuple[int, Card]] = []
        for card in valid_cards:
            if card.name == "Curse":
                # prioritize giving opponents curses first
                prioritized_cards.append((1, card))
            elif CardType.Attack in card.type:
                # lower priority for giving opponents attack cards
                prioritized_cards.append((3, card))
            elif CardType.Victory in card.type:
                # lower priority for giving opponents victory cards
                prioritized_cards.append((4, card))
            else:
                # prioritize giving all other cards second
                prioritized_cards.append((2, card))
        prioritized_cards.sort(key=lambda x: x[0])

        return prioritized_cards[0][1]

    @overload
    def torturer(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        num_discard: int = -1,
        options: Literal[True] = True,
        discard: Literal[False] = False,
    ) -> int:
        ...

    @overload
    def torturer(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        num_discard: int = -1,
        options: Literal[False] = False,
        discard: Literal[True] = True,
    ) -> List[Card]:
        ...

    def torturer(
        self,
        player: "Player",
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        num_discard: int = -1,
        options: bool = False,
        discard: bool = False,
    ) -> Union[int, List[Card]]:
        if options:
            if game.supply.pile_length("Curse") == 0:
                return Torturer.Choice.GainCurse

            discard_cards = self.get_optional_discard(player.hand.cards, player)
            max_discard = min(2, len(player.hand))
            if len(discard_cards) >= max_discard:
                return Torturer.Choice.Discard

            return Torturer.Choice.GainCurse

        if discard:
            assert valid_cards is not None
            assert num_discard >= 0
            discard_cards = self.sort_for_discard(
                valid_cards,
                player.state.actions,
                player,
                game,
            )
            return discard_cards[:num_discard]

        raise InvalidBotImplementation(
            "Either options or discard must be true when playing torturer"
        )

    def trading_post(
        self,
        player: "Player",
        game: "Game",
        valid_cards: List[Card],
    ) -> List[Card]:
        trash_cards = self.sort_for_trash(valid_cards, player, game)
        return trash_cards[:2]

    def upgrade(
        self,
        player: "Player",
        game: "Game",
        valid_cards: List[Card],
        trash: bool = False,
        gain: bool = False,
    ) -> Card:
        if trash:
            min_price_card = min(valid_cards, key=lambda card: card.get_cost(player, game))
            return min_price_card

        if gain:
            max_price_card = max(valid_cards, key=lambda card: card.get_cost(player, game))
            return max_price_card

        raise InvalidBotImplementation(
            "Either trash or gain must be true when playing upgrade"
        )

    def wishing_well(
        self,
        player: "Player",
        game: "Game",
    ) -> Card:
        return copper


class OptimizedBot(Bot):
    """
    Bot with logic for playing and reacting to all cards.

    """

    def __init__(
        self,
        decider: Optional[Decider] = None,
        player_id: str = "bot",
    ):
        decider = decider if decider else OptimizedBotDecider()
        super().__init__(decider=decider, player_id=player_id)



import logging
import random
from typing import List, Optional

from pyminion.core import CardType, Card, Deck, DeckCounter, Pile, Supply, Trash
from pyminion.exceptions import InvalidGameSetup, InvalidPlayerCount
from pyminion.expansions.base import (copper, curse, duchy, estate, gold,
                                      province, silver)
from pyminion.player import Player
from pyminion.result import GameOutcome, GameResult, PlayerSummary

logger = logging.getLogger()


class Game:
    """
    Core class the emulates a game of Dominion

    Attributes:
        players: List of players in the game.
        expansions: List expansions (and their cards) eligible to be used in the game's supply.
        kingdom_cards: Specify any specific cards to be used in the supply.
        start_deck: List of cards each player will start the game with. Default = [7 Coppers + 3 Estates].
        random_order: If True, scrambles the order of players (to offset first player advantage).
        log_stdout: If True, logs game to stdout.
        log_file: If True, logs game to log file.
        log_file_name: Name of the file to be logged to. Default = "game.log"

    """

    def __init__(
        self,
        players: List[Player],
        expansions: List[List[Card]],
        kingdom_cards: Optional[List[Card]] = None,
        start_deck: Optional[List[Card]] = None,
        random_order: bool = True,
        log_stdout: bool = True,
        log_file: bool = False,
        log_file_name: str = "game.log",
    ):

        if len(players) < 1:
            raise InvalidPlayerCount("Game must have at least one player")
        if len(players) > 4:
            raise InvalidPlayerCount("Game can have at most four players")
        self.players = players
        self.expansions = expansions
        self.kingdom_cards = kingdom_cards
        self.all_game_cards: List[Card] = []
        self.card_cost_reduction = 0
        self.start_deck = start_deck
        self.random_order = random_order
        self.trash = Trash()

        if log_stdout:
            # Set up a handler that logs to stdout
            c_handler = logging.StreamHandler()
            c_handler.setLevel(logging.INFO)
            c_format = logging.Formatter("%(message)s")
            c_handler.setFormatter(c_format)
            logger.addHandler(c_handler)

        if log_file:
            # Set up a handler that dumps the log to a file
            f_handler = logging.FileHandler(log_file_name, mode="w")
            f_handler.setLevel(logging.INFO)
            f_format = logging.Formatter("%(message)s")
            f_handler.setFormatter(f_format)
            logger.addHandler(f_handler)

    def _create_basic_piles(self) -> List[Pile]:
        """
        Create the basic piles that are applicable to almost all games of Dominion.

        """

        if len(self.players) == 1:
            VICTORY_LENGTH = 5
            CURSE_LENGTH = 10
            COPPER_LENGTH = 53

        elif len(self.players) == 2:
            VICTORY_LENGTH = 8
            CURSE_LENGTH = 10
            COPPER_LENGTH = 46

        elif len(self.players) == 3:
            VICTORY_LENGTH = 12
            CURSE_LENGTH = 20
            COPPER_LENGTH = 39

        else:
            VICTORY_LENGTH = 12
            CURSE_LENGTH = 30
            COPPER_LENGTH = 32

        SILVER_LENGTH = 40
        GOLD_LENGTH = 30

        basic_cards = [
            copper,
            silver,
            gold,
            estate,
            duchy,
            province,
            curse,
        ]

        basic_piles = []
        for card in basic_cards:
            if card.name == "Copper":
                basic_piles.append(Pile([card] * COPPER_LENGTH))
            elif card.name == "Silver":
                basic_piles.append(Pile([card] * SILVER_LENGTH))
            elif card.name == "Gold":
                basic_piles.append(Pile([card] * GOLD_LENGTH))
            elif CardType.Victory in card.type:
                basic_piles.append(Pile([card] * VICTORY_LENGTH))
            elif CardType.Curse in card.type:
                basic_piles.append(Pile([card] * CURSE_LENGTH))

        return basic_piles

    def _create_kingdom_piles(self) -> List[Pile]:
        """
        Create the kingdom piles that vary from kingdom to kingdom.
        This should be 10 piles each with 10 cards.

        """
        PILE_LENGTH: int = 10
        KINGDOM_PILES: int = 10

        # All avaliable cards from chosen expansions
        kingdom_options = [card for expansion in self.expansions for card in expansion]

        # If user chooses kingdom cards, put them in the supply
        if self.kingdom_cards:
            if invalid_cards := [
                card for card in self.kingdom_cards if card not in kingdom_options
            ]:
                raise InvalidGameSetup(
                    f"Invalid game setup: {invalid_cards} not in provided expansions"
                )
            chosen_cards = len(self.kingdom_cards)

        else:
            chosen_cards = 0

        chosen_piles = (
            [Pile([card] * PILE_LENGTH) for card in self.kingdom_cards]
            if chosen_cards
            else []
        )
        # The rest of the supply is random cards
        if chosen_cards:
            for card in self.kingdom_cards:
                kingdom_options.remove(card)  # Do not duplicate any user chosen cards
        kingdom_ten = random.sample(kingdom_options, KINGDOM_PILES - chosen_cards)
        random_piles = [Pile([card] * PILE_LENGTH) for card in kingdom_ten]

        return chosen_piles + random_piles

    def _create_supply(self) -> Supply:
        """
        Create a supply consisting of basic cards
        avaliable in every kingdom as well
        as the kingdom specific cards.

        """

        basic_piles = self._create_basic_piles()
        kingdom_piles = self._create_kingdom_piles()
        all_piles = basic_piles + kingdom_piles
        self.all_game_cards = [pile.cards[0] for pile in all_piles]
        return Supply(all_piles)

    def start(self) -> None:
        logger.info("\nStarting Game...\n")

        self.supply = self._create_supply()
        logger.info(f"Supply: \n{self.supply}")

        if self.random_order:
            random.shuffle(self.players)
        if not self.start_deck:
            self.start_deck = [copper] * 7 + [estate] * 3

        for player in self.players:
            player.reset()
            player.discard_pile = Deck(self.start_deck[:])
            logger.info(f"\n{player} starts with {player.discard_pile}")
            player.draw(5)

    def is_over(self) -> bool:
        """
        The game is over if any 3 supply piles are empty or
        if the province pile is empty.

        Return True if the game is over

        """
        empty_piles: int = 0
        for pile in self.supply.piles:

            # are provinces empty?
            if pile.name == "Province" and len(pile) == 0:
                return True

            # are three piles empty?
            if len(pile) == 0:
                empty_piles += 1
                if empty_piles >= 3:
                    return True

        return False

    def play(self) -> GameResult:
        self.start()
        while True:
            for player in self.players:
                player.take_turn(self)

                # reset card cost reduction
                self.card_cost_reduction = 0

                if self.is_over():
                    result = self.summerize_game()
                    logging.info(f"\n{result}")
                    return result

    def get_winners(self) -> List[Player]:
        """
        The player with the most victory points wins.
        If the highest scores are tied at the end of the game,
        the tied player who has had the fewest turns wins the game.
        If the tied players have had the same number of turns, they tie.

        Returns a list of players. If there is a single player in
        the list, that is the sole winner. If there are multiple players
        in the list, they are have tied for first.

        """
        # if one player only, they win by default
        if len(self.players) == 1:
            return [self.players[0]]

        # temporarily set first player as winner
        high_score = self.players[0].get_victory_points()
        winners = [self.players[0]]

        # iterate the rest of the players in the game
        for player in self.players[1:]:
            score = player.get_victory_points()

            # if this player scored more,
            # mark them as winner and high score
            if score > high_score:
                high_score = score
                winners = [player]

            # if scores are equal
            elif score == high_score:

                # players tie if number of turns is equal
                if player.turns == winners[0].turns:
                    winners.append(player)

                # otherwise, player with fewer turns wins
                elif player.turns < winners[0].turns:
                    winners = [player]

                # note
                # we can compare to just the first player in winners,
                # because if there were multiple players in winners
                # they would have equal score and turns

        return winners

    def summerize_game(self) -> GameResult:
        """
        Called at the end of the game,
        this creates a summary of the game

        """

        player_summaries = []
        winners = self.get_winners()

        for order, player in enumerate(self.players):

            # player won
            if player in winners and len(winners) == 1:
                result = GameOutcome.win

            # player tied
            elif player in winners:
                result = GameOutcome.tie

            # player lost
            else:
                result = GameOutcome.loss

            summary = PlayerSummary(
                player=player,
                result=result,
                score=player.get_victory_points(),
                turns=player.turns,
                shuffles=player.shuffles,
                turn_order=order + 1,
                deck=DeckCounter(player.get_all_cards()),
            )
            player_summaries.append(summary)

        game_result = GameResult(
            game=self,
            turns=winners[0].turns,
            winners=winners,
            player_summaries=player_summaries,
        )
        return game_result



class InsufficientActions(Exception):
    """
    Player does not have remaining actions required to play card

    """


class InsufficientMoney(Exception):
    """
    Player does not have enough money to buy card

    """


class InsufficientBuys(Exception):
    """
    Player does not have enough buys to purchase card

    """


class EmptyPile(Exception):
    """
    The pile does is empty

    """


class PileNotFound(Exception):
    """
    The pile does not exist

    """


class CardNotFound(Exception):
    """
    The card does not exist

    """


class InvalidInput(Exception):
    """
    Base class for input exceptions

    """


class InvalidCardPlay(Exception):
    """
    Unable to play card

    """


class InvalidBinaryInput(InvalidInput):
    """
    Invalid response, valid choices are "y" or "n"

    """


class InvalidSingleCardInput(InvalidInput):
    """
    Invalid response, that input cannot be discarded

    """


class InvalidMultiCardInput(InvalidInput):
    """
    Invalid response, that input cannot be discarded

    """


class InvalidMultiOptionInput(InvalidInput):
    """
    Invalid response, not a valid option

    """


class InvalidDeckPositionInput(InvalidInput):
    """
    Invalid deck position response

    """


class InvalidPlayerCount(Exception):
    """
    Invalid number of players in game

    """


class InvalidGameSetup(Exception):
    """
    Invalid game setup

    """


class InvalidBotImplementation(Exception):
    """
    Invalid game setup

    """



import logging
from typing import Iterator

from pyminion.bots.bot import Bot, BotDecider
from pyminion.core import Card
from pyminion.expansions.base import gold, province, silver
from pyminion.player import Player
from pyminion.game import Game

logger = logging.getLogger()


class BigMoneyDecider(BotDecider):
    """
    Only buys money and provinces

    """

    def buy_priority(self, player: Player, game: Game) -> Iterator[Card]:
        money = player.state.money
        if money >= 8:
            yield province
        if money >= 6:
            yield gold
        if money >= 3:
            yield silver


class BigMoney(Bot):
    def __init__(
        self,
        player_id: str = "big_money",
    ):
        super().__init__(decider=BigMoneyDecider(), player_id=player_id)



# Specific Bot Implementations
from pyminion.bots.examples.bandit_bot import BanditBot
from pyminion.bots.examples.big_money import BigMoney
from pyminion.bots.examples.big_money_smithy import BigMoneySmithy
from pyminion.bots.examples.big_money_ultimate import BigMoneyUltimate
from pyminion.bots.examples.chapel_bot import ChapelBot



import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, List, Optional

from pyminion.core import (AbstractDeck, Action, CardType, Card, Deck, DiscardPile, Hand,
                           Playmat, Supply, Trash)
from pyminion.decider import Decider
from pyminion.exceptions import (CardNotFound, EmptyPile, InsufficientBuys,
                                 InsufficientMoney, InvalidCardPlay)

if TYPE_CHECKING:
    from pyminion.game import Game


logger = logging.getLogger()


@dataclass
class State:
    """
    Hold state during a player's turn

    """

    actions: int = 1
    money: int = 0
    buys: int = 1


class Player:
    """
    Basic representation of a player including the piles of cards they own
    and the basic actions they can take to manipulate the state of the game.

    """

    def __init__(
        self,
        decider: Decider,
        deck: Optional[Deck] = None,
        discard_pile: Optional[DiscardPile] = None,
        hand: Optional[Hand] = None,
        playmat: Optional[Playmat] = None,
        state: Optional[State] = None,
        player_id: str = "",
    ):
        self.decider = decider
        self.deck = deck if deck else Deck()
        self.discard_pile = discard_pile if discard_pile else DiscardPile()
        self.hand = hand if hand else Hand()
        self.playmat = playmat if playmat else Playmat()
        self.state = state if state else State()
        self.player_id = player_id
        self.turns: int = 0
        self.shuffles: int = 0
        self.actions_played_this_turn: int = 0

    def __repr__(self):
        return f"{self.player_id}"

    def reset(self):
        """
        Reset the state of the player to a pre-game state.
        Required for resetting player deck and state between games when running simulations.

        """
        self.turns = 0
        self.shuffles = 0
        self.actions_played_this_turn = 0
        self.deck.cards = []
        self.discard_pile.cards = []
        self.hand.cards = []

    def draw(
        self,
        num_cards: int = 1,
        destination: Optional[AbstractDeck] = None,
        silent: bool = False,
    ) -> None:
        """
        Draw cards from deck and add them to the specified destination.
        Defaults to drawing one card and adding to the player's hand.

        By default, drawn cards and deck shuffles are logged.
        Call with silent = True to disable logging of drawn cards

        """
        if destination is None:
            destination = self.hand
        drawn_cards: AbstractDeck = AbstractDeck()
        for _ in range(num_cards):
            # Both deck and discard empty -> do nothing
            if len(self.discard_pile) == 0 and len(self.deck) == 0:
                pass
            # Deck is empty -> shuffle discard pile into deck
            elif len(self.deck) == 0:
                logger.info(f"{self} shuffles their deck")
                self.discard_pile.move_to(self.deck)
                self.deck.shuffle()
                self.shuffles += 1
                draw_card = self.deck.draw()
                destination.add(draw_card)
                drawn_cards.add(draw_card)
            else:
                draw_card = self.deck.draw()
                destination.add(draw_card)
                drawn_cards.add(draw_card)
        if not silent:
            logger.info(f"{self} draws {drawn_cards}")

    def discard(self, target_card: Card) -> None:
        """
        Move specified card from the player's hand to the player's discard pile.

        """
        for card in self.hand.cards:
            if card == target_card:
                self.discard_pile.add(self.hand.remove(card))
                logger.info(f"{self} discards {card}")
                return

    def play(self, target_card: Card, game: "Game", generic_play: bool = True) -> None:
        """
        Find target card in player's hand and play it.

        If generic_play is true, card is moved from player's hand to playmat
        and player action count decreased by 1. This is the default behavior
        but is overridden for cards like vassal and throne room.

        """
        if target_card not in self.hand.cards:
            raise CardNotFound(f"Invalid play, {target_card} not in hand")
        for card in self.hand.cards:
            if card.name == target_card.name:
                if CardType.Action in card.type:
                    self.actions_played_this_turn += 1
                    card.play(player=self, game=game, generic_play=generic_play)
                    return
                if CardType.Treasure in card.type:
                    card.play(player=self, game=game)
                    return
        raise InvalidCardPlay(f"Invalid play, {target_card} could not be played")

    def exact_play(self, card: Card, game: "Game", generic_play: bool = True) -> None:
        """
        Similar to previous play method, except exact card to play must be specified.
        This is method is necessary when playing cards not in the player's hand, such as vassal.

        """
        if CardType.Action in card.type:
            self.actions_played_this_turn += 1
            card.play(player=self, game=game, generic_play=generic_play)
        elif CardType.Treasure in card.type:
            card.play(player=self, game=game)
        else:
            raise InvalidCardPlay(f"Unable to play {card} with type {card.type}")

    def multi_play(self, card: Card, game: "Game", state: Any, generic_play: bool = True) -> Any:
        """
        Similar to previous exact_play method, except card's multi_play method is called.
        This is method is necessary when playing "Throne Room variants".

        """
        if CardType.Action in card.type:
            assert isinstance(card, Action)
            self.actions_played_this_turn += 1
            state = card.multi_play(player=self, game=game, state=state, generic_play=generic_play)
        else:
            raise InvalidCardPlay(f"Unable to play {card} with type {card.type}")
        return state

    def buy(self, card: Card, game: "Game") -> None:
        """
        Buy a card from the supply and add to player's discard pile.
        Check that player has sufficient money and buys to gain the card.

        """
        assert isinstance(card, Card)
        if card.get_cost(self, game) > self.state.money:
            raise InsufficientMoney(
                f"{self.player_id}: Not enough money to buy {card.name}"
            )
        if self.state.buys < 1:
            raise InsufficientBuys(
                f"{self.player_id}: Not enough buys to buy {card.name}"
            )
        try:
            game.supply.gain_card(card)
        except EmptyPile as e:
            raise e
        self.state.money -= card.get_cost(self, game)
        self.state.buys -= 1
        self.discard_pile.add(card)
        logger.info(f"{self} buys {card}")

    def gain(
        self, card: Card, supply: "Supply", destination: Optional[AbstractDeck] = None
    ) -> None:
        """
        Gain a card from the supply and add to destination.
        Defaults to adding the card to the player's discard pile.

        """
        if destination is None:
            destination = self.discard_pile

        gain_card = supply.gain_card(card)
        destination.add(gain_card)
        logger.info(f"{self} gains {gain_card}")

    def trash(
        self, target_card: Card, trash: "Trash", source: Optional[AbstractDeck] = None
    ) -> None:
        """
        Move a card from source to the trash.
        Defaults to moving the card from the player's hand.

        """
        if source is None:
            source = self.hand

        for card in source.cards:
            if card == target_card:
                trash.add(source.remove(card))
                logger.info(f"{self} trashes {card}")

                break

    def start_turn(self) -> None:
        """
        Increase turn counter and reset state

        """
        self.turns += 1
        self.actions_played_this_turn = 0
        self.state.actions = 1
        self.state.money = 0
        self.state.buys = 1

    def start_action_phase(self, game: "Game") -> None:
        while self.state.actions > 0:
            logger.info(f"{self.player_id}'s hand: {self.hand}")

            viable_actions = [card for card in self.hand.cards if CardType.Action in card.type]
            if not viable_actions:
                return

            card = self.decider.action_phase_decision(viable_actions, self, game)
            if card is None:
                return

            self.play(card, game)

    def start_treasure_phase(self, game: "Game") -> None:
        viable_treasures = [card for card in self.hand.cards if CardType.Treasure in card.type]
        while len(viable_treasures) > 0:
            logger.info(f"Hand: {self.hand}")

            cards = self.decider.treasure_phase_decision(viable_treasures, self, game)
            if len(cards) == 0:
                break

            for card in cards:
                self.exact_play(card, game)
            cards_str = ", ".join([str(c) for c in cards])
            logger.info(f"{self.player_id} played {cards_str}")

            viable_treasures = [card for card in self.hand.cards if CardType.Treasure in card.type]

    def start_buy_phase(self, game: "Game") -> None:
        while self.state.buys > 0:
            logger.info(f"\nSupply:{game.supply}")
            logger.info(f"Money: {self.state.money}")
            logger.info(f"Buys: {self.state.buys}")

            valid_cards = [
                c
                for c in game.supply.avaliable_cards()
                if c.get_cost(self, game) <= self.state.money
            ]
            card = self.decider.buy_phase_decision(
                valid_cards=valid_cards,
                player=self,
                game=game,
            )

            if card is None:
                logger.info(f"{self} buys nothing")
                break

            self.buy(card, game)

    def start_cleanup_phase(self) -> None:
        """
        Move hand and playmat cards into discard pile and draw 5 new cards.

        """
        self.discard_pile.cards += self.hand.cards
        self.discard_pile.cards += self.playmat.cards
        self.hand.cards = []
        self.playmat.cards = []
        self.draw(5)
        self.state.actions = 1
        self.state.money = 0
        self.state.buys = 1

    def take_turn(self, game: "Game") -> None:
        self.start_turn()
        logger.info(f"\nTurn {self.turns} - {self.player_id}")
        self.start_action_phase(game)
        self.start_treasure_phase(game)
        self.start_buy_phase(game)
        self.start_cleanup_phase()

    def get_all_cards(self) -> List[Card]:
        """
        Get a list of all the cards the player has in their possession.

        """
        all_cards = (
            self.deck.cards
            + self.discard_pile.cards
            + self.playmat.cards
            + self.hand.cards
        )
        return all_cards

    def get_card_count(self, card: Card) -> int:
        """
        Get count of a specific card in player's whole deck.

        """
        return self.get_all_cards().count(card)

    def get_victory_points(self) -> int:
        """
        Return the number of victory points a player has.

        """
        total_vp: int = 0
        for card in self.get_all_cards():
            if CardType.Victory in card.type or CardType.Curse in card.type:
                total_vp += card.score(self)
        return total_vp

    def get_treasure_money(self) -> int:
        """
        Return the amount of money a player has in their deck from treasure cards.

        """
        total_money: int = 0
        for card in self.get_all_cards():
            if CardType.Treasure in card.type:
                total_money += card.money
        return total_money

    def get_action_money(self) -> int:
        """
        Return the amount of money a player has in their deck from action cards.

        """
        total_money: int = 0
        for card in self.get_all_cards():
            if CardType.Action in card.type:
                total_money += card.money
        return total_money

    def get_deck_money(self) -> int:
        """
        Return total count of all money a player has in their deck.

        """
        treasure_money = self.get_treasure_money()
        action_money = self.get_action_money()
        return treasure_money + action_money

    def is_attacked(self, player: "Player", attack_card: Card, game: "Game") -> bool:
        attacked = True
        for card in self.hand.cards:
            if card.name == "Moat":
                block = self.decider.binary_decision(
                    prompt=f"Would you like to block {player.player_id}'s {attack_card} with your Moat? y/n: ",
                    card=card,
                    player=self,
                    game=game,
                    relevant_cards=[attack_card],
                )
                attacked &= not block
            elif card.name == "Diplomat":
                card.on_attack(self, attack_card, game)

        return attacked



import logging
from typing import Iterator

from pyminion.bots.bot import Bot, BotDecider
from pyminion.core import CardType, Card
from pyminion.expansions.base import duchy, estate, gold, province, silver, smithy
from pyminion.player import Player
from pyminion.game import Game

logger = logging.getLogger()


class BigMoneyUltimateDecider(BotDecider):
    """
    Attempt the following buys in order:

    Buy Province if total money in deck is > 15
    Buy Duchy if remaining Provinces < 5
    Buy Estate if remaining Provinces < 3
    Buy Gold
    Buy Duchy if remaining Provinces < 7
    Buy Smithy if #Smithies < (#Treasures / 11)
    Buy Silver

    """

    def action_priority(self, player: "Player", game: "Game") -> Iterator[Card]:
        yield smithy

    def buy_priority(self, player: "Player", game: "Game") -> Iterator["Card"]:
        money = player.state.money
        deck_money = player.get_deck_money()
        num_province = game.supply.pile_length(pile_name="Province")
        num_smithy = player.get_card_count(card=smithy)
        num_treasure = len(
            [card for card in player.get_all_cards() if CardType.Treasure in card.type]
        )

        if deck_money > 15 and money >= 8:
            yield province
        if num_province < 5 and money >= 5:
            yield duchy
        if num_province < 3 and money >= 2:
            yield estate
        if money >= 6:
            yield gold
        if num_province < 7 and money >= 5:
            yield duchy
        if num_smithy < num_treasure / 11 and money >= 4:
            yield smithy
        if money >= 3:
            yield silver


class BigMoneyUltimate(Bot):
    def __init__(
        self,
        player_id: str = "big_money_ultimate",
    ):
        super().__init__(decider=BigMoneyUltimateDecider(), player_id=player_id)



from typing import Iterator

from pyminion.bots.optimized_bot import OptimizedBot, OptimizedBotDecider
from pyminion.core import CardType, Card
from pyminion.expansions.base import (
    bandit,
    duchy,
    estate,
    gold,
    province,
    silver,
    smithy,
)
from pyminion.player import Player
from pyminion.game import Game


class BanditBotDecider(OptimizedBotDecider):
    """
    Attempt the following buys in order:

    Buy Province if total money in deck is > 15
    Buy Duchy if remaining Provinces < 5
    Buy Estate if remaining Provinces < 3
    Buy Gold
    Buy Duchy if remaining Provinces < 7
    Buy Bandit if #Bandits < 1
    Buy Smithy if #Smithies < (#Treasures / 11)
    Buy Silver

    """

    def action_priority(self, player: Player, game: Game) -> Iterator[Card]:
        yield bandit
        yield smithy

    def buy_priority(self, player: Player, game: Game) -> Iterator[Card]:

        money = player.state.money
        deck_money = player.get_deck_money()
        num_province = game.supply.pile_length(pile_name="Province")
        num_smithy = player.get_card_count(card=smithy)
        num_bandit = player.get_card_count(card=bandit)
        num_treasure = len(
            [card for card in player.get_all_cards() if CardType.Treasure in card.type]
        )

        if deck_money > 15 and money >= 8:
            yield province
        if num_province < 5 and money >= 5:
            yield duchy
        if num_province < 3 and money >= 2:
            yield estate
        if money >= 6:
            yield gold
        if num_province < 7 and money >= 5:
            yield duchy
        if num_bandit < 1 and money >= 5:
            yield bandit
        if num_smithy < num_treasure / 11 and money >= 4:
            yield smithy
        if money >= 3:
            yield silver


class BanditBot(OptimizedBot):
    def __init__(
        self,
        player_id: str = "bandit_bot",
    ):
        super().__init__(decider=BanditBotDecider(), player_id=player_id)



from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from pyminion.core import DeckCounter
    from pyminion.game import Game
    from pyminion.player import Player


class GameOutcome(Enum):
    """
    player can either lose, tie, or win the game

    """

    loss = -1
    tie = 0
    win = 1


@dataclass
class PlayerSummary:
    """
    holds summary of a player from a complete game

    """

    player: "Player"
    result: GameOutcome
    score: int
    turns: int
    shuffles: int
    turn_order: int
    deck: "DeckCounter"

    def __repr__(self):

        order_format = "None"
        if self.turn_order == 1:
            order_format = "1st"

        elif self.turn_order == 2:
            order_format = "2nd"

        elif self.turn_order == 3:
            order_format = "3rd"

        elif self.turn_order == 4:
            order_format = "4th"

        player = f"Player: {self.player.player_id}"
        result = f"Result: {self.result.name}"
        score = f"Score: {self.score}"
        turns = f"Turns: {self.turns}"
        shuffles = f"Shuffles: {self.shuffles}"
        order = f"Turn Order: {order_format}"
        deck = f"Deck: {self.deck}"

        return f"{player}\n{result}\n{score}\n{turns}\n{shuffles}\n{order}\n{deck}"


@dataclass
class GameResult:
    """
    holds summary of a complete game

    """

    game: "Game"
    winners: List["Player"]
    turns: int
    player_summaries: List[PlayerSummary]

    def __repr__(self):
        if len(self.winners) == 1:
            result = f"{self.winners[0]} won in {self.turns} turns"
        else:
            result = f"{[w for w in self.winners]} tied after {self.turns} turns"

        format_summaries = ""
        for s in self.player_summaries:
            format_summaries += f"\n{s}"

        return f"Game Result: {result}{format_summaries}"


@dataclass
class PlayerSimulatorResult:
    player: "Player"
    wins: int
    losses: int
    ties: int


@dataclass
class SimulatorResult:
    """
    holds summary of game outcomes over a simulation

    """

    iterations: int
    game_results: List[GameResult]
    player_results: List[PlayerSimulatorResult]

    def __repr__(self):
        title = f"ran {self.iterations} games"

        format_results = ""
        for result in self.player_results:
            format_results += f"\n{result.player.player_id} won {result.wins}, lost {result.losses}, tied {result.ties}"

        return f"Simulation Result: {title}{format_results}"



from typing import TYPE_CHECKING, List, Optional, Protocol

if TYPE_CHECKING:
    from pyminion.core import Card, Player
    from pyminion.game import Game


class Decider(Protocol):
    """
    Interface for prompting a player for a decision.

    """

    def action_phase_decision(
        self,
        valid_actions: List["Card"],
        player: "Player",
        game: "Game",
    ) -> Optional["Card"]:
        raise NotImplementedError("action_phase_decision is not implemented")

    def treasure_phase_decision(
        self,
        valid_treasures: List["Card"],
        player: "Player",
        game: "Game",
    ) -> List["Card"]:
        raise NotImplementedError("treasure_phase_decision is not implemented")

    def buy_phase_decision(
        self,
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
    ) -> Optional["Card"]:
        raise NotImplementedError("buy_phase_decision is not implemented")

    def binary_decision(
        self,
        prompt: str,
        card: "Card",
        player: "Player",
        game: "Game",
        relevant_cards: Optional[List["Card"]] = None,
    ) -> bool:
        raise NotImplementedError("binary_decision is not implemented")

    def multiple_option_decision(
        self,
        card: "Card",
        options: List[str],
        player: "Player",
        game: "Game",
        num_choices: int = 1,
        unique: bool = True,
    ) -> List[int]:
        raise NotImplementedError("multiple_option_decision is not implemented")

    def discard_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_discard: int = 0,
        max_num_discard: int = -1,
    ) -> List["Card"]:
        raise NotImplementedError("discard_decision is not implemented")

    def trash_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_trash: int = 0,
        max_num_trash: int = -1,
    ) -> List["Card"]:
        raise NotImplementedError("trash_decision is not implemented")

    def gain_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_gain: int = 0,
        max_num_gain: int = -1,
    ) -> List["Card"]:
        raise NotImplementedError("gain_decision is not implemented")

    def topdeck_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_topdeck: int = 0,
        max_num_topdeck: int = -1,
    ) -> List["Card"]:
        raise NotImplementedError("topdeck_decision is not implemented")

    def deck_position_decision(
        self,
        prompt: str,
        card: "Card",
        player: "Player",
        game: "Game",
        num_deck_cards: int,
    ) -> int:
        raise NotImplementedError("deck_position_decision is not implemented")

    def reveal_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_reveal: int = 0,
        max_num_reveal: int = -1,
    ) -> List["Card"]:
        raise NotImplementedError("reveal_decision is not implemented")

    def pass_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_pass: int = 0,
        max_num_pass: int = -1,
    ) -> List["Card"]:
        raise NotImplementedError("pass_decision is not implemented")

    def name_card_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        min_num_name: int = 0,
        max_num_name: int = -1,
    ) -> List["Card"]:
        raise NotImplementedError("name_card_decision is not implemented")

    def multi_play_decision(
        self,
        prompt: str,
        card: "Card",
        valid_cards: List["Card"],
        player: "Player",
        game: "Game",
        required: bool = True,
    ) -> Optional["Card"]:
        raise NotImplementedError("multi_play_decision is not implemented")



from enum import IntEnum, unique
import logging
from typing import TYPE_CHECKING, Any, List

from pyminion.core import AbstractDeck, Action, Card, CardType, Treasure, Victory
from pyminion.player import Player
from pyminion.exceptions import EmptyPile
from pyminion.expansions.base import curse, duchy, estate, gold, silver

if TYPE_CHECKING:
    from pyminion.game import Game


logger = logging.getLogger()


class Baron(Action):
    """
    +1 Buy

    You may discard an Estate for +4 money. If you don't, gain an Estate.

    """

    @unique
    class Choice(IntEnum):
        DiscardEstate = 0
        GainEstate = 1

    def __init__(self):
        super().__init__(name="Baron", cost=4, type=(CardType.Action,))

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")
        if generic_play:
            super().generic_play(player)

        player.state.buys += 1

        discard_estate = False
        if estate in player.hand.cards:
            options = [
                "Discard estate for +4 money",
                "Gain an estate",
            ]
            responses = player.decider.multiple_option_decision(self, options, player, game)
            assert len(responses) == 1
            response = responses[0]

            discard_estate = (response == Baron.Choice.DiscardEstate)

        if discard_estate:
            player.discard(estate)
            player.state.money += 4
        elif game.supply.pile_length(estate.name) > 0:
            player.gain(estate, game.supply)


class Bridge(Action):
    """
    +1 buy, +$1

    This turn, cards (everywhere) cost $1 less.

    """

    def __init__(self):
        super().__init__(name="Bridge", cost=4, type=(CardType.Action,), money=1)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")
        if generic_play:
            super().generic_play(player)

        player.state.buys += 1
        player.state.money += 1

        game.card_cost_reduction += 1


class Conspirator(Action):
    """
    +2 Money

    If you've played 3 or more Actions this turn (counting this), +1 Card and +1 Action.

    """

    def __init__(self):
        super().__init__(name="Conspirator", cost=4, type=(CardType.Action,), money=2)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")
        if generic_play:
            super().generic_play(player)

        player.state.money += 2

        if player.actions_played_this_turn >= 3:
            player.draw(1)
            player.state.actions += 1


class Courtier(Action):
    """
    Reveal a card from your hand. For each type it has (Action, Attack, etc.),
    choose one: +1 Action; or +1 Buy; or +3 Money; or gain a Gold. The choices
    must be different.

    """

    @unique
    class Choice(IntEnum):
        Action = 0
        Buy = 1
        Money = 2
        GainGold = 3

    def __init__(self):
        super().__init__(name="Courtier", cost=5, type=(CardType.Action,))

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")
        if generic_play:
            super().generic_play(player)

        hand_len = len(player.hand)

        if hand_len == 0:
            return

        if hand_len == 1:
            reveal_card = player.hand.cards[0]
        else:
            reveal_cards = player.decider.reveal_decision(
                prompt="Reveal a card from your hand: ",
                card=self,
                valid_cards=player.hand.cards,
                player=player,
                game=game,
                min_num_reveal = 1,
                max_num_reveal = 1,
            )
            assert len(reveal_cards) == 1
            reveal_card = reveal_cards[0]

        logger.info(f"{player} reveals {reveal_card}")

        num_choices = min(len(reveal_card.type), 4)

        if num_choices == 4:
            choices = [c.value for c in Courtier.Choice]
        else:
            options = [
                "+1 Action",
                "+1 Buy",
                "+3 Money",
                "Gain a Gold",
            ]
            choices = player.decider.multiple_option_decision(
                card=self,
                options=options,
                player=player,
                game=game,
                num_choices=num_choices,
                unique=True,
            )
            assert len(set(choices)) == num_choices # ensure number of unique choices is correct

        for choice in choices:
            if choice == Courtier.Choice.Action:
                player.state.actions += 1
            elif choice == Courtier.Choice.Buy:
                player.state.buys += 1
            elif choice == Courtier.Choice.Money:
                player.state.money += 3
            elif choice == Courtier.Choice.GainGold:
                player.gain(gold, game.supply)
            else:
                raise ValueError(f"Unknown courtier choice '{choice}'")


class Courtyard(Action):
    """
    +3 Cards

    Put a card from your hand onto your deck.

    """

    def __init__(self):
        super().__init__(name="Courtyard", cost=2, type=(CardType.Action,), draw=3)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")
        if generic_play:
            super().generic_play(player)

        player.draw(3)

        topdeck_cards = player.decider.topdeck_decision(
            prompt="Enter the card you would like to topdeck: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
            min_num_topdeck=1,
            max_num_topdeck=1,
        )
        assert len(topdeck_cards) == 1
        topdeck_card = topdeck_cards[0]

        player.hand.remove(topdeck_card)
        player.deck.add(topdeck_card)


class Diplomat(Action):
    """
    +2 Cards

    If you have 5 or fewer cards in hand (after drawing), +2 Actions.

    When another player plays an Attack card, you may first reveal this from a
    hand of 5 or more cards, to draw 2 cards then discard 3.

    """

    def __init__(self):
        super().__init__("Diplomat", 4, (CardType.Action, CardType.Reaction), draw=2)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")
        if generic_play:
            super().generic_play(player)

        player.draw(2)
        if len(player.hand) <= 5:
            player.state.actions += 2

    def on_attack(self, player: "Player", attack_card: Card, game: "Game") -> None:
        if len(player.hand) < 5:
            return

        reveal = player.decider.binary_decision(
            prompt=f"Reveal {self} to draw 2 cards then discard 3? y/n: ",
            card=self,
            player=player,
            game=game,
            relevant_cards=[attack_card],
        )
        if not reveal:
            return

        logger.info(f"{player} reveals {self}")

        player.draw(2)

        discard_cards = player.decider.discard_decision(
            prompt="Enter the cards you would like to discard: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
            min_num_discard=3,
            max_num_discard=3,
        )
        assert len(discard_cards) == 3

        for card in discard_cards:
            player.discard(card)


class Duke(Victory):
    """
    Worth 1VP per Duchy you have.

    """

    def __init__(self):
        super().__init__("Duke", 5, (CardType.Victory,))

    def score(self, player: Player) -> int:
        vp = 0
        for card in player.get_all_cards():
            if card.name == duchy.name:
                vp += 1
        return vp


class Harem(Treasure, Victory):
    def __init__(self):
        Treasure.__init__(
            self,
            name="Harem",
            cost=6,
            type=(CardType.Treasure, CardType.Victory),
            money=2,
        )

    def play(self, player: Player, game: "Game") -> None:
        player.playmat.add(self)
        player.hand.remove(self)
        player.state.money += self.money

    def score(self, player: Player) -> int:
        vp = 2
        return vp


class Ironworks(Action):
    """
    Gain a card costing up to $4.
    If the gained card is an...

    Action card, +1 Action
    Treasure card, +$1
    Victory card, +1 Card

    """

    def __init__(self):
        super().__init__(name="Ironworks", cost=4, type=(CardType.Action,))

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")
        if generic_play:
            super().generic_play(player)

        gain_cards = player.decider.gain_decision(
            prompt="Gain a card costing up to 4 money: ",
            card=self,
            valid_cards=[
                card for card in game.supply.avaliable_cards() if card.get_cost(player, game) <= 4
            ],
            player=player,
            game=game,
            min_num_gain=1,
            max_num_gain=1,
        )
        assert len(gain_cards) == 1
        gain_card = gain_cards[0]
        assert gain_card.get_cost(player, game) <= 4

        player.gain(card=gain_card, supply=game.supply)

        if CardType.Action in gain_card.type:
            player.state.actions += 1

        if CardType.Treasure in gain_card.type:
            player.state.money += 1

        if CardType.Victory in gain_card.type:
            player.draw(1)


class Lurker(Action):
    """
    +1 Action

    Choose one: Trash an Action card from the Supply, or gain an Action card from the trash.

    """

    @unique
    class Choice(IntEnum):
        TrashAction = 0
        GainAction = 1

    def __init__(self):
        super().__init__(name="Lurker", cost=2, type=(CardType.Action,), actions=1)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")
        if generic_play:
            super().generic_play(player)

        player.state.actions += 1

        supply_action_cards = [
            c for c in game.supply.avaliable_cards() if CardType.Action in c.type
        ]
        trash_action_cards = [
            c for c in game.trash.cards if CardType.Action in c.type
        ]

        if len(supply_action_cards) == 0 and len(trash_action_cards) == 0:
            return

        options = [
            "Trash an Action card from the Supply",
            "Gain an Action card from the trash",
        ]
        choices = player.decider.multiple_option_decision(
            card=self,
            options=options,
            player=player,
            game=game,
        )
        assert len(choices) == 1
        choice = choices[0]

        if choice == Lurker.Choice.TrashAction:
            if len(supply_action_cards) == 0:
                return

            trash_cards = player.decider.trash_decision(
                prompt="Choose a card from the Supply to trash",
                card=self,
                valid_cards=supply_action_cards,
                player=player,
                game=game,
                min_num_trash=1,
                max_num_trash=1,
            )
            assert len(trash_cards) == 1
            trash_card = trash_cards[0]

            game.supply.trash_card(trash_card, game.trash)

        elif choice == Lurker.Choice.GainAction:
            if len(trash_action_cards) == 0:
                return

            gain_cards = player.decider.gain_decision(
                prompt="Choose a card to gain from the trash",
                card=self,
                valid_cards=trash_action_cards,
                player=player,
                game=game,
                min_num_gain=1,
                max_num_gain=1,
            )
            assert len(gain_cards) == 1
            gain_card = gain_cards[0]

            game.trash.remove(gain_card)
            player.discard_pile.add(gain_card)

        else:
            raise ValueError(f"Unknown lurker choice '{choice}'")


class Masquerade(Action):
    """
    +2 Cards

    Each player with any cards in hand passes one to the next such player to
    their left, at once. Then you may trash a card from your hand.

    """

    def __init__(self):
        super().__init__(name="Masquerade", cost=3, type=(CardType.Action,), draw=2)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(2)

        # get players who have at least 1 card in their hand
        valid_players = [p for p in game.players if len(p.hand) > 0]

        # prompt each player to choose a card to pass
        passed_cards: List[Card] = []
        for p in valid_players:
            pass_cards = player.decider.pass_decision(
                prompt="Pick a card to pass to the player on your left: ",
                card=self,
                valid_cards=p.hand.cards,
                player=player,
                game=game,
                min_num_pass=1,
                max_num_pass=1,
            )
            assert len(pass_cards) == 1
            pass_card = pass_cards[0]

            passed_cards.append(pass_card)

        # pass the cards
        for idx, p in enumerate(valid_players):
            c = passed_cards[idx]
            p.hand.remove(c)
            next_idx = (idx + 1) % len(valid_players)
            next_player = valid_players[next_idx]
            next_player.hand.add(c)
            logger.info(f"{p} passes {c} to {next_player}")

        if len(player.hand) == 0:
            return

        trash = player.decider.binary_decision(
            prompt="Would you like to trash a card from your hand?",
            card=self,
            player=player,
            game=game,
        )

        if trash:
            trash_cards = player.decider.trash_decision(
                prompt="Choose a card from your hand to trash",
                card=self,
                valid_cards=player.hand.cards,
                player=player,
                game=game,
                min_num_trash=1,
                max_num_trash=1,
            )
            assert len(trash_cards) == 1
            trash_card = trash_cards[0]

            player.trash(trash_card, game.trash)


class Mill(Action, Victory):
    """
    +1 card, +1 action

    You may discard 2 cards, for +$2.

    """

    def __init__(self):
        Action.__init__(self, "Mill", 4, (CardType.Action, CardType.Victory), actions=1, draw=1)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(1)
        player.state.actions += 1

        discard = player.decider.binary_decision(
            prompt="Do you want to discard 2 cards for +2 money?",
            card=self,
            player=player,
            game=game,
        )

        if discard:
            hand_len = len(player.hand)
            if hand_len <= 2:
                discard_cards = player.hand.cards[:]
            else:
                discard_cards = player.decider.discard_decision(
                    prompt="Enter the cards you would like to discard: ",
                    card=self,
                    valid_cards=player.hand.cards,
                    player=player,
                    game=game,
                    min_num_discard=2,
                    max_num_discard=2,
                )
                assert len(discard_cards) == 2

            if len(discard_cards) == 2:
                player.state.money += 2

            for c in discard_cards:
                player.discard(c)

    def score(self, player: Player) -> int:
        vp = 1
        return vp


class MiningVillage(Action):
    """
    +1 card, +2 actions

    You may trash this for +$2.

    """

    def __init__(self):
        super().__init__(name="Mining Village", cost=4, type=(CardType.Action,), actions=2, draw=1)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        self._play(player, game, None, generic_play)

    def multi_play(
        self, player: Player, game: "Game", state: Any, generic_play: bool = True
    ) -> Any:
        return self._play(player, game, state, generic_play)

    def _play(
        self, player: Player, game: "Game", state: Any, generic_play: bool = True
    ) -> Any:
        player.draw(1)
        player.state.actions += 2

        trashed = False if state is None else bool(state)

        # if the card has not previously been trashed, give the player the
        # option to trash it now
        if not trashed:
            trashed = player.decider.binary_decision(
                prompt="Trash Mining Village for +2 money? ",
                card=self,
                player=player,
                game=game,
            )

            if trashed:
                player.state.money += 2
                player.trash(self, game.trash, source=player.playmat)

        return trashed


class Minion(Action):
    """
    +1 action

    Choose one: +$2; or discard your hand, +4 Cards, and each other player
    with at least 5 cards in hand discards their hand and draws 4 cards.

    """

    @unique
    class Choice(IntEnum):
        Money = 0
        DiscardDrawAttack = 1

    def __init__(self):
        super().__init__(name="Minion", cost=5, type=(CardType.Action, CardType.Attack), actions=1)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 1

        # opponents react to the card before the choice is made
        is_attacked: List[bool] = []
        for opponent in game.players:
            if opponent is not player:
                ret = opponent.is_attacked(player, self, game)
                is_attacked.append(ret)

        options = [
            "+2 Money",
            "Discard your hand, draw 4 cards, and each other player with at least 5 cards discards their hand and draws 4 cards",
        ]
        choices = player.decider.multiple_option_decision(
            card=self,
            options=options,
            player=player,
            game=game,
        )
        assert len(choices) == 1
        choice = choices[0]

        if choice == Minion.Choice.Money:
            player.state.money += 2
        elif choice == Minion.Choice.DiscardDrawAttack:
            for _ in range(len(player.hand.cards)):
                player.discard(player.hand.cards[0])
            player.draw(4)

            i = 0
            for opponent in game.players:
                if opponent is not player and is_attacked[i]:
                    if len(opponent.hand) >= 5:
                        for _ in range(len(opponent.hand.cards)):
                            opponent.discard(opponent.hand.cards[0])
                        opponent.draw(4)
                    i += 1
        else:
            raise ValueError(f"Unknown minion choice '{choice}'")


class Nobles(Action, Victory):
    """
    Choose one: +3 Cards; or +2 Actions.

    """

    @unique
    class Choice(IntEnum):
        Cards = 0
        Actions = 1

    def __init__(self):
        Action.__init__(self, "Nobles", 6, (CardType.Action, CardType.Victory))

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        options = [
            "+3 Cards",
            "+2 Actions",
        ]
        choices = player.decider.multiple_option_decision(
            card=self,
            options=options,
            player=player,
            game=game,
        )
        assert len(choices) == 1
        choice = choices[0]

        if choice == Nobles.Choice.Cards:
            player.draw(3)
        elif choice == Nobles.Choice.Actions:
            player.state.actions += 2
        else:
            raise ValueError(f"Unknown nobles choice '{choice}'")

    def score(self, player: Player) -> int:
        vp = 2
        return vp


class Patrol(Action):
    """
    +3 cards

    Reveal the top 4 cards of your deck. Put the Victory cards and Curses into
    your hand. Put the rest back in any order.

    """

    def __init__(self):
        super().__init__(name="Patrol", cost=5, type=(CardType.Action,), draw=3)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(3)

        revealed = AbstractDeck()
        player.draw(num_cards=4, destination=revealed, silent=True)
        logger.info(f"{player} reveals {revealed}")

        victory_curse_cards = [
            card
            for card in revealed.cards
            if CardType.Curse in card.type or CardType.Victory in card.type
        ]

        for card in victory_curse_cards:
            player.hand.add(revealed.remove(card))

        num_topdeck = len(revealed.cards)
        if num_topdeck <= 1:
            topdeck_cards = revealed.cards
        else:
            topdeck_cards = player.decider.topdeck_decision(
                prompt="Enter the cards in the order you would like to topdeck: ",
                card=self,
                valid_cards=revealed.cards,
                player=player,
                game=game,
                min_num_topdeck=num_topdeck,
                max_num_topdeck=num_topdeck,
            )

        for card in topdeck_cards:
            player.deck.add(card)


class Pawn(Action):
    """
    Choose two: +1 Card; +1 Action; +1 Buy; +1 Money.
    The choices must be different.

    """

    @unique
    class Choice(IntEnum):
        Card = 0
        Action = 1
        Buy = 2
        Money = 3

    def __init__(self):
        super().__init__(name="Pawn", cost=2, type=(CardType.Action,))

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        options = [
            "+1 Card",
            "+1 Action",
            "+1 Buy",
            "+1 Money",
        ]
        choices = player.decider.multiple_option_decision(
            card=self,
            options=options,
            player=player,
            game=game,
            num_choices=2,
            unique=True,
        )
        assert len(set(choices)) == 2 # ensure number of unique choices is correct

        for choice in choices:
            if choice == Pawn.Choice.Card:
                player.draw(1)
            elif choice == Pawn.Choice.Action:
                player.state.actions += 1
            elif choice == Pawn.Choice.Buy:
                player.state.buys += 1
            elif choice == Pawn.Choice.Money:
                player.state.money += 1
            else:
                raise ValueError(f"Unknown pawn choice '{choice}'")


class Replace(Action):
    """
    Trash a card from your hand. Gain a card costing up to $2 more than it. If
    the gained card is an Action or Treasure, put it onto your deck; if it's a
    Victory card, each other player gains a Curse.

    """

    def __init__(self):
        super().__init__(name="Replace", cost=5, type=(CardType.Action, CardType.Attack))

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        trash_cards = player.decider.trash_decision(
            prompt="Trash a card form your hand: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
            min_num_trash=1,
            max_num_trash=1,
        )
        assert len(trash_cards) == 1
        trash_card = trash_cards[0]

        max_cost = trash_card.get_cost(player, game) + 2
        gain_cards = player.decider.gain_decision(
            prompt=f"Gain a card costing up to {max_cost} money: ",
            card=self,
            valid_cards=[
                card
                for card in game.supply.avaliable_cards()
                if card.get_cost(player, game) <= max_cost
            ],
            player=player,
            game=game,
            min_num_gain=1,
            max_num_gain=1,
        )
        assert len(gain_cards) == 1
        gain_card = gain_cards[0]
        assert gain_card.get_cost(player, game) <= max_cost

        player.trash(trash_card, trash=game.trash)

        if CardType.Action in gain_card.type or CardType.Treasure in gain_card.type:
            player.gain(gain_card, game.supply, destination=player.deck)
        else:
            player.gain(gain_card, game.supply)

        if CardType.Victory in gain_card.type:
            for opponent in game.players:
                if opponent is not player and opponent.is_attacked(player, self, game):
                    # attempt to gain a curse. if curse pile is empty, proceed
                    try:
                        opponent.gain(curse, game.supply)
                    except EmptyPile:
                        pass


class SecretPassage(Action):
    """
    +2 cards
    +1 action

    Take a card from your hand and put it anywhere in your deck.

    """

    def __init__(self):
        super().__init__(name="Secret Passage", cost=4, type=(CardType.Action,), actions=1, draw=2)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(2)
        player.state.actions += 1

        insert_cards = player.decider.topdeck_decision(
            prompt="Enter the card you would like to insert in your deck: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
            min_num_topdeck=1,
            max_num_topdeck=1,
        )
        assert len(insert_cards) == 1
        insert_card = insert_cards[0]

        len_deck = len(player.deck)
        if len_deck == 0:
            index = 0
        else:
            index = player.decider.deck_position_decision(
                prompt=f"Enter the deck position (bottom = 1, top = {len_deck+1}): ",
                card=self,
                player=player,
                game=game,
                num_deck_cards=len_deck,
            )
            assert 0 <= index <= len_deck

        player.hand.remove(insert_card)
        player.deck.cards.insert(index, insert_card)


class ShantyTown(Action):
    """
    +2 Actions

    Reveal your hand.
    If you have no Action cards in hand, +2 Cards.

    """

    def __init__(self):
        super().__init__(name="Shanty Town", cost=3, type=(CardType.Action,), actions=2)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 2

        if not any(CardType.Action in c.type for c in player.hand.cards):
            player.draw(2)


class Steward(Action):
    """
    Choose one: +2 Cards; or +2 money; or trash 2 cards from your hand.

    """

    @unique
    class Choice(IntEnum):
        Cards = 0
        Money = 1
        Trash = 2

    def __init__(self):
        super().__init__(name="Steward", cost=3, type=(CardType.Action,))

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        options = [
            "+2 Cards",
            "+2 Money",
            "Trash 2 cards from your hand",
        ]
        choices = player.decider.multiple_option_decision(
            card=self,
            options=options,
            player=player,
            game=game,
        )
        assert len(choices) == 1
        choice = choices[0]

        if choice == Steward.Choice.Cards:
            player.draw(2)
        elif choice == Steward.Choice.Money:
            player.state.money += 2
        elif choice == Steward.Choice.Trash:
            trash_cards = self._get_trash_cards(player, game)
            for card in trash_cards:
                player.trash(card, game.trash)
        else:
            raise ValueError(f"Unknown steward choice '{choice}'")

    def _get_trash_cards(self, player: Player, game: "Game") -> List[Card]:

        if len(player.hand) <= 2:
            return player.hand.cards[:]

        trash_cards = player.decider.trash_decision(
            prompt="Enter 2 cards you would like to trash from your hand: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
            min_num_trash=2,
            max_num_trash=2,
        )
        assert len(trash_cards) == 2

        return trash_cards


class Swindler(Action):
    """
    +2 Money

    Each other player trashes the top card of their deck and gains a card with
    the same cost that you choose.

    """

    def __init__(self):
        super().__init__(name="Swindler", cost=3, type=(CardType.Action, CardType.Attack), money=2)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.money += 2

        for opponent in game.players:
            if opponent is not player and opponent.is_attacked(player, self, game):
                revealed_cards = AbstractDeck()
                opponent.draw(num_cards=1, destination=revealed_cards, silent=True)
                trashed_card = revealed_cards.cards[0]
                game.trash.add(trashed_card)
                trashed_cost = trashed_card.get_cost(player, game)

                logger.info(f"{opponent} trashes {trashed_card}")

                valid_cards = [
                    c
                    for c in game.supply.avaliable_cards()
                    if c.get_cost(player, game) == trashed_cost
                ]
                if len(valid_cards) == 0:
                    continue

                gain_cards = player.decider.gain_decision(
                    prompt=f"Pick a cost {trashed_cost} card for {opponent} to gain: ",
                    card=self,
                    valid_cards=valid_cards,
                    player=player,
                    game=game,
                    min_num_gain=1,
                    max_num_gain=1,
                )
                assert len(gain_cards) == 1
                gain_card = gain_cards[0]
                opponent.discard_pile.add(gain_card)

                logger.info(f"{opponent} gains {gain_card}")


class Torturer(Action):
    """
    +3 cards

    Each other player either discards 2 cards or gains a Curse to their
    hand, their choice. (They may pick an option they can't do.)

    """

    @unique
    class Choice(IntEnum):
        Discard = 0
        GainCurse = 1

    def __init__(self):
        super().__init__(name="Torturer", cost=5, type=(CardType.Action, CardType.Attack), draw=3)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(3)

        for opponent in game.players:
            if opponent is not player and opponent.is_attacked(player, self, game):
                options = [
                    "Discard 2 cards",
                    "Gain a Curse to your hand",
                ]
                choices = opponent.decider.multiple_option_decision(
                    card=self,
                    options=options,
                    player=opponent,
                    game=game,
                )
                assert len(choices) == 1
                choice = choices[0]

                if choice == Torturer.Choice.Discard:
                    self._discard(opponent, game)
                elif choice == Torturer.Choice.GainCurse:
                    self._gain_curse(opponent, game)
                else:
                    raise ValueError(f"Unknown torturer choice '{choice}'")

    def _discard(self, opponent: Player, game: "Game") -> None:
        num_discard = min(2, len(opponent.hand))
        if num_discard == 0:
            return

        discard_cards = opponent.decider.discard_decision(
            prompt=f"You must discard {num_discard} card(s) from your hand: ",
            card=self,
            valid_cards=opponent.hand.cards,
            player=opponent,
            game=game,
            min_num_discard=num_discard,
            max_num_discard=num_discard,
        )
        assert len(discard_cards) == num_discard

        for card in discard_cards:
            opponent.discard(target_card=card)

    def _gain_curse(self, opponent: Player, game: "Game") -> None:
        try:
            opponent.gain(
                card=curse,
                supply=game.supply,
                destination=opponent.hand,
            )
        except EmptyPile:
            pass


class TradingPost(Action):
    """
    Trash 2 cards from your hand. If you did, gain a Silver to your hand.

    """

    def __init__(self):
        super().__init__(name="Trading Post", cost=5, type=(CardType.Action,))

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        len_hand = len(player.hand)
        if len_hand == 0:
            return
        elif len_hand <= 2:
            trash_cards = player.hand.cards[:]
        else:
            trash_cards = player.decider.trash_decision(
                prompt="Trash 2 cards from your hand: ",
                card=self,
                valid_cards=player.hand.cards,
                player=player,
                game=game,
                min_num_trash=2,
                max_num_trash=2,
            )
            assert len(trash_cards) == 2

        for card in trash_cards:
            player.trash(card, game.trash)

        if len(trash_cards) == 2:
            # attempt to gain a silver to player's hand.
            # if silver pile is empty, proceed
            try:
                player.gain(silver, game.supply, player.hand)
            except EmptyPile:
                pass


class Upgrade(Action):
    """
    +1 card, +1 action
    Trash a card from your hand. Gain a card costing exactly $1 more than it.

    """

    def __init__(self):
        super().__init__(name="Upgrade", cost=5, type=(CardType.Action,), actions=1, draw=1)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(1)
        player.state.actions += 1

        trash_cards = player.decider.trash_decision(
            prompt="Trash a card form your hand: ",
            card=self,
            valid_cards=player.hand.cards,
            player=player,
            game=game,
            min_num_trash=1,
            max_num_trash=1,
        )
        assert len(trash_cards) == 1
        trash_card = trash_cards[0]

        player.trash(trash_card, trash=game.trash)

        new_cost = trash_card.get_cost(player, game) + 1
        valid_cards = [
            card
            for card in game.supply.avaliable_cards()
            if card.get_cost(player, game) == new_cost
        ]

        if len(valid_cards) == 0:
            return

        gain_cards = player.decider.gain_decision(
            prompt=f"Gain a card costing {new_cost} money: ",
            card=self,
            valid_cards=valid_cards,
            player=player,
            game=game,
            min_num_gain=1,
            max_num_gain=1,
        )
        assert len(gain_cards) == 1
        gain_card = gain_cards[0]
        assert gain_card.get_cost(player, game) == new_cost

        player.gain(gain_card, game.supply)


class WishingWell(Action):
    """
    +1 card, +1 action

    Name a card, then reveal the top card of your deck. If you named it,
    put it into your hand.

    """

    def __init__(self):
        super().__init__(name="Wishing Well", cost=3, type=(CardType.Action,), actions=1, draw=1)

    def play(
        self, player: Player, game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(1)
        player.state.actions += 1

        named_cards = player.decider.name_card_decision(
            prompt="Name a card: ",
            card=self,
            valid_cards=game.all_game_cards,
            player=player,
            game=game,
            min_num_name=1,
            max_num_name=1,
        )
        assert len(named_cards) == 1
        name = named_cards[0].name

        logger.info(f"{player} names {name}")

        revealed = AbstractDeck()
        player.draw(1, revealed, silent=True)
        revealed_card = revealed.cards[0]
        revealed_name = revealed_card.name

        msg = f"{player} reveals {revealed_name} and "
        if revealed_name == name:
            player.hand.add(revealed_card)
            msg += "puts it into their hand"
        else:
            player.deck.add(revealed_card)
            msg += "topdecks it"

        logger.info(msg)


baron = Baron()
bridge = Bridge()
conspirator = Conspirator()
courtier = Courtier()
courtyard = Courtyard()
diplomat = Diplomat()
duke = Duke()
harem = Harem()
ironworks = Ironworks()
lurker = Lurker()
masquerade = Masquerade()
mill = Mill()
mining_village = MiningVillage()
minion = Minion()
nobles = Nobles()
patrol = Patrol()
pawn = Pawn()
replace = Replace()
secret_passage = SecretPassage()
shanty_town = ShantyTown()
steward = Steward()
swindler = Swindler()
torturer = Torturer()
trading_post = TradingPost()
upgrade = Upgrade()
wishing_well = WishingWell()


intrigue_set: List[Card] = [
    baron,
    bridge,
    conspirator,
    courtier,
    courtyard,
    diplomat,
    duke,
    harem,
    ironworks,
    lurker,
    masquerade,
    mill,
    mining_village,
    minion,
    nobles,
    patrol,
    pawn,
    replace,
    secret_passage,
    shanty_town,
    steward,
    swindler,
    torturer,
    trading_post,
    upgrade,
    wishing_well,
]



from typing import Iterator

from pyminion.bots.optimized_bot import OptimizedBot, OptimizedBotDecider
from pyminion.core import Card
from pyminion.expansions.base import chapel, duchy, estate, gold, province, silver
from pyminion.player import Player
from pyminion.game import Game

class ChapelBotDecider(OptimizedBotDecider):
    """
    Attempt the following buys in order:

    Buy Chapel if #Chapel < 1
    Buy Province if total money in deck is > 15
    Buy Duchy if remaining Provinces < 5
    Buy Estate if remaining Provinces < 3
    Buy Gold
    Buy Duchy if remaining Provinces < 7
    Buy Silver

    """

    def action_priority(self, player: Player, game: Game) -> Iterator[Card]:
        yield chapel

    def buy_priority(self, player: Player, game: Game) -> Iterator[Card]:

        money = player.state.money
        deck_money = player.get_deck_money()
        num_province = game.supply.pile_length(pile_name="Province")
        num_chapel = player.get_card_count(card=chapel)

        if num_chapel < 1 and money >= 2:
            yield chapel
        if deck_money > 15 and money >= 8:
            yield province
        if num_province < 5 and money >= 5:
            yield duchy
        if num_province < 3 and money >= 2:
            yield estate
        if money >= 6:
            yield gold
        if num_province < 7 and money >= 5:
            yield duchy
        if money >= 3:
            yield silver


class ChapelBot(OptimizedBot):
    def __init__(
        self,
        player_id: str = "chapel_bot",
    ):
        super().__init__(decider=ChapelBotDecider(), player_id=player_id)
